from __future__ import annotations

import copy
import logging
from typing import Optional, Literal, Tuple

import pandas as pd
from matplotlib.axes import Axes

import okama.portfolios.core as core
import okama.portfolios.mc as mc
import okama.portfolios.cashflow_strategies as cf
import okama.portfolios.dcf_calculations as dcf_calculations
from okama import settings
from okama.common import validators
from okama.common.helpers import helpers
from okama.common.solver import Result

logger = logging.getLogger(__name__)


# Constants for cashflow strategy types
class CashflowStrategyType:
    """Constants for cashflow strategy types."""

    FIXED_AMOUNT = "fixed_amount"
    FIXED_PERCENTAGE = "fixed_percentage"


# Constants for withdrawal goals
class WithdrawalGoal:
    """Constants for withdrawal goals."""

    MAINTAIN_BALANCE_PV = "maintain_balance_pv"
    MAINTAIN_BALANCE_FV = "maintain_balance_fv"
    SURVIVAL_PERIOD = "survival_period"


class PortfolioDCF:
    """
    Discounted cash flow (DCF) methods for a `Portfolio`.

    This class is available via `Portfolio.dcf`.

    Parameters
    ----------
    parent : Portfolio
        Parent `Portfolio` instance.
    discount_rate : float or None, default None
        Annual effective discount rate used to calculate present values (PV).
        If None and the portfolio has inflation data, the geometric mean of inflation is used.
        Otherwise, `settings.DEFAULT_DISCOUNT_RATE` is used.

    Examples
    --------
    >>> pf = ok.Portfolio(first_date="2015-01", last_date="2024-10")
    >>> pf.dcf.wealth_index(discounting="fv").head()
    """

    def __init__(
        self,
        parent: core.Portfolio,
        discount_rate: Optional[float] = None,
    ):
        self.parent = parent
        self.discount_rate = discount_rate
        self._wealth_index_fv = pd.DataFrame(dtype=float)
        self._monte_carlo_wealth_fv = pd.DataFrame(dtype=float)
        self._monte_carlo_cash_flow_fv = pd.DataFrame(dtype=float)
        self._cash_flow_fv = pd.Series(dtype=float, name="cash_flow_fv")
        self.mc = mc.MonteCarlo(self)
        self._cashflow_parameters: Optional[cf.CashFlow] = cf.IndexationStrategy(parent=parent)

    def __repr__(self):
        dic = {
            "Portfolio symbol": self.parent.symbol,
            "Monte Carlo distribution": self.mc.distribution,
            "Monte Carlo period": self.mc.period,
            "Cash flow strategy": self.cashflow_parameters.NAME if hasattr(self.cashflow_parameters, "NAME") else None,
            "discount_rate": self.discount_rate,
        }
        return repr(pd.Series(dic))

    @property
    def discount_rate(self) -> float:
        """
        Annual effective discount rate for portfolio cash flow.

        Returns
        -------
        float
            Cash flow discount rate.
        """
        return self._discount_rate

    @discount_rate.setter
    def discount_rate(self, discount_rate: Optional[float]):
        self._wealth_index_fv = pd.DataFrame()
        self._monte_carlo_wealth_fv = pd.DataFrame()
        if discount_rate is None and hasattr(self.parent, "inflation"):
            self._discount_rate = helpers.Frame.get_cagr(self.parent.inflation_ts)
            if self._discount_rate is None:
                # if time series is too short (<12 month) CAGR is not defined
                self._discount_rate = settings.DEFAULT_DISCOUNT_RATE
        elif discount_rate is None:  # and not hasattr(self.parent, "inflation")
            self._discount_rate = settings.DEFAULT_DISCOUNT_RATE
        else:
            validators.validate_real("discount rate", discount_rate)
            self._discount_rate = discount_rate

    @property
    def cashflow_parameters(self) -> Optional[cf.CashFlow]:
        """
        Return the cash flow strategy parameters.

        Returns
        -------
        CashFlow or None
            The cash flow strategy parameters (CashFlow instance) or None if not defined.
        """
        return self._cashflow_parameters

    @cashflow_parameters.setter
    def cashflow_parameters(self, cashflow_parameters: Optional[cf.CashFlow]):
        if isinstance(cashflow_parameters, cf.CashFlow):
            self._cashflow_parameters = cashflow_parameters
            self.cashflow_parameters._clear_cf_cache()
        elif cashflow_parameters is None:
            self._cashflow_parameters = cashflow_parameters
        else:
            raise TypeError("cashflow_parameters must be a CashFlow instance or None")

    def set_mc_parameters(
        self,
        distribution: str = "norm",
        distribution_parameters: Optional[tuple] = None,
        period: int = 1,
        mc_number: int = 100,
    ):
        """
        Add Monte Carlo simulation parameters to PortfolioDCF.

        Parameters
        ----------
        distribution : {'norm', 'lognorm', 't'}, default 'norm'
            Distribution used to generate random rates of return:

            - 'norm' for normal distribution
            - 'lognorm' for lognormal distribution
            - 't' for Student's t-distribution

        distribution_parameters : tuple or None, default None
            Distribution parameters to generate random rate of return.
            (mean, standard deviation) for normal distribution.
            (shape, loc, scale) for lognormal distribution.
            (df, loc, scale) for Student distribution.
            Use None for any element to let SciPy estimate it via the distribution `.fit()` method
            (for example, `(3, None, None)` for Student's t-distribution with df=3 and auto loc/scale).

        period : int, default 1
            Forecast period for portfolio wealth index time series (in years).

        mc_number : int, default 100
            Number of random wealth indexes to generate with Monte Carlo simulation.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(first_date="2015-01", last_date="2024-10")  # create Portfolio with default parameters
        >>> # Set Monte Carlo parameters
        >>> pf.dcf.set_mc_parameters(distribution="lognorm", period=10, mc_number=100)
        >>> # Set the cash flow strategy. It's required to generate random wealth indexes.
        >>> ind = ok.IndexationStrategy(pf) # create IndexationStrategy linked to the portfolio
        >>> ind.initial_investment = 10_000  # add initial investments size
        >>> ind.frequency = "year"  # set cash flow frequency
        >>> ind.amount = -1_500  # set withdrawal size
        >>> ind.indexation = "inflation"
        >>> # Assign the strategy to Portfolio
        >>> pf.dcf.cashflow_parameters = ind
        >>> # Plot wealth index with cash flow
        >>> pf.dcf.wealth_index(discounting="fv", include_negative_values=False).plot()
        >>> plt.show()
        """
        self.mc.distribution = distribution
        self.mc.distribution_parameters = distribution_parameters
        self.mc.period = period
        self.mc.mc_number = mc_number

    def wealth_index(self, discounting: Literal["fv", "pv"], include_negative_values: bool = False) -> pd.DataFrame:
        """
        Calculate wealth index time series for the portfolio with cash flow (contributions and withdrawals)
        using historical rate of returns.

        Wealth index (Cumulative Wealth Index) is a time series that presents the value of portfolio over
        historical time period considering cash flows.

        Accumulated inflation time series is added if `inflation=True` in the Portfolio.

        If there is no cash flow, Wealth index is obtained from the accumulated return multiplied
        by the initial investments. That is: initial_investment * (Accumulated_Return + 1)

        Parameters
        ----------
        discounting : {'fv', 'pv'}
            Type of discounting to apply:
            - 'fv': Future Values - nominal values without discounting
            - 'pv': Present Values - values discounted to present using the discount rate
        include_negative_values : bool, default False
            Whether to include negative wealth index values in the result.
            If False, negative values are replaced with zeros.

        Returns
        -------
        pd.DataFrame
            Time series of wealth index values for portfolio and accumulated inflation (if applicable).

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['VOO.US', 'GLD.US'], weights=[0.8, 0.2])
        >>> ind = ok.IndexationStrategy(pf)  # Set Cash Flow Strategy parameters
        >>> ind.initial_investment = 100  # initial investments value
        >>> ind.frequency = "year"  # withdrawals frequency
        >>> ind.amount = -6  # initial withdrawals amount
        >>> ind.indexation = "inflation"  # the indexation is equal to inflation
        >>> pf.dcf.cashflow_parameters = ind  # assign the strategy to Portfolio
        >>> pf.dcf.wealth_index(discounting="fv", include_negative_values=False).plot()
        >>> plt.show()
        """
        if self.cashflow_parameters is None:
            raise AttributeError("'cashflow_parameters' is not defined.")
        if self._wealth_index_fv.empty:
            df = self.parent._add_inflation()
            infl_symbol = self.parent.inflation if hasattr(self.parent, "inflation") else None
            df = dcf_calculations.get_wealth_indexes_fv_with_cashflow(
                ror=df,
                portfolio_symbol=self.parent.symbol,
                inflation_symbol=infl_symbol,
                cashflow_parameters=self.cashflow_parameters,
                task="backtest",
            )
            self._wealth_index_fv = self.parent._make_df_if_series(df)
        wealth_index_fv = self._wealth_index_fv.copy()
        if not include_negative_values:
            wealth_index_fv_s = dcf_calculations.remove_negative_values(self._wealth_index_fv[self.parent.name])
            wealth_index_fv[self.parent.name] = wealth_index_fv_s.fillna(0)
        if discounting.lower() == "fv":
            return wealth_index_fv
        elif discounting.lower() == "pv":
            return dcf_calculations.discount_monthly_cash_flow(wealth_index_fv, self.discount_rate, reverse=True)
        else:
            raise ValueError("'discounting' must be either 'fv' or 'pv'")

    def cash_flow_ts(self, discounting: Literal["fv", "pv"], remove_if_wealth_index_negative: bool = True) -> pd.Series:
        """
        Cash flow Future Values (FV) or Present Values (PV) time series for the portfolio.

        The cash flow time series presents the values of contributions and withdrawals over
        historical time period.

        Parameters
        ----------
        discounting : {'fv', 'pv'}
            Type of discounting to apply:
            - 'fv': Future Values - nominal values without discounting
            - 'pv': Present Values - values discounted to present using the discount rate
        remove_if_wealth_index_negative : bool, default True
            If True, cash flow values are replaced with 0 when wealth index is negative (or zero).

        Returns
        -------
        pd.Series
            Time series of cash flow values for portfolio.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['VOO.US', 'GLD.US'], weights=[0.8, 0.2])
        >>> ind = ok.IndexationStrategy(pf)  # Set Cash Flow Strategy parameters
        >>> ind.initial_investment = 100  # initial investments value
        >>> ind.frequency = "year"  # withdrawals frequency
        >>> ind.amount = -0.5 * 12  # initial withdrawals amount
        >>> ind.indexation = "inflation"  # the indexation is equal to inflation
        >>> pf.dcf.cashflow_parameters = ind  # assign the strategy to Portfolio
        >>> pf.dcf.cash_flow_ts(discounting="fv", remove_if_wealth_index_negative=True).plot()
        >>> plt.show()
        """
        if self.cashflow_parameters is None:
            raise AttributeError("'cashflow_parameters' is not defined.")
        if self._cash_flow_fv.empty:
            df = self.parent.ror
            self._cash_flow_fv = dcf_calculations.get_cash_flow_fv(
                ror=df,
                portfolio_symbol=self.parent.symbol,
                cashflow_parameters=self.cashflow_parameters,
                task="backtest",
            )
        cash_flow_fv = self._cash_flow_fv.copy()
        if remove_if_wealth_index_negative:
            wealth_index = self.wealth_index(discounting="fv", include_negative_values=False)
            condition = wealth_index[self.parent.name] == 0
            cash_flow_fv[condition] = 0
        if discounting.lower() == "fv":
            return cash_flow_fv
        elif discounting.lower() == "pv":
            return dcf_calculations.discount_monthly_cash_flow(cash_flow_fv, self.discount_rate, reverse=True)
        else:
            raise ValueError("'discounting' must be either 'fv' or 'pv'")

    @property
    def wealth_index_fv_with_assets(self) -> pd.DataFrame:
        """
        Wealth index Future Values (FV) time series for the portfolio and all assets considering cash flow (contributions and
        withdrawals).

        Wealth index (Cumulative Wealth Index) is a time series that presents the value of portfolio over
        historical time period. Accumulated inflation time series is added if `inflation=True` in the Portfolio.

        If there is no cash flow, Wealth index is obtained from the accumulated return multiplied
        by the initial investments. That is: initial_investment * (Accumulated_Return + 1)

        Returns
        -------
        DataFrame
            Time series of wealth index values for portfolio, each asset and accumulated inflation.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['VOO.US', 'GLD.US'], weights=[0.8, 0.2])
        >>> ind = ok.IndexationStrategy(pf)  # Set Cash Flow Strategy parameters
        >>> ind.initial_investment = 100  # initial investments value
        >>> ind.frequency = "year"  # withdrawals frequency
        >>> ind.amount = -0.5 * 12  # initial withdrawals amount
        >>> ind.indexation = "inflation"  # the indexation is equal to inflation
        >>> pf.dcf.cashflow_parameters = ind  # assign the strategy to Portfolio
        >>> pf.dcf.wealth_index_fv_with_assets.plot()
        >>> plt.show()
        """
        ls = [self.parent.ror, self.parent.assets_ror]
        ror_df = pd.concat(ls, axis=1, join="inner", copy="false")
        wealth_df = ror_df.apply(
            dcf_calculations.get_wealth_indexes_fv_with_cashflow,
            axis=0,
            args=(None, None, self.cashflow_parameters, "backtest"),  # symbol  # inflation_symbol
        )
        if hasattr(self.parent, "inflation"):
            inflation_wi = helpers.Frame.get_wealth_indexes(
                ror=self.parent.inflation_ts, initial_amount=self.cashflow_parameters.initial_investment
            )
            wealth_df = pd.concat([wealth_df, inflation_wi], axis=1, join="inner", copy="false")
        return wealth_df

    def survival_period_hist(self, threshold: float = 0) -> float:
        """
        Calculate the period when the portfolio has positive balance considering withdrawals on the historical data.

        The portfolio survival period (longevity period) depends on the investment strategy: asset allocation,
        rebalancing, withdrawals rate etc.

        Parameters
        ----------
        threshold : float, default 0
            The percentage of the initial investments when the portfolio balance considered voided.
            This parameter is important to use in cash flow strategies with a fixed
            withdrawal percentage (PercentageStrategy).

        Returns
        -------
        float
            The portfolio survival period (longevity period) in years.

        Examples
        --------
        >>> pf = ok.Portfolio(["SPY.US", "AGG.US"], ccy="USD", first_date="2010-01", last_date="2024-10")
        >>> # set cash flow strategy
        >>> ind = ok.IndexationStrategy(pf)  # create cash flow strategy linked to the portfolio
        >>> ind.initial_investment = 10_000  # add initial investment to cash flow strategy
        >>> ind.amount = -2_500  # set annual withdrawal size
        >>> ind.frequency = "year"  # set withdrawal frequency to year
        >>> pf.dcf.cashflow_parameters = ind
        >>> # Calculate the historical survival period for the cash flow strategy.
        >>> # The balance is considered voided when it's equal to 0 (threshold=0)
        >>> pf.dcf.survival_period_hist(threshold=0)
        5.1
        """
        return helpers.Date.get_period_length(
            last_date=self.survival_date_hist(threshold=threshold), first_date=self.parent.first_date
        )

    def survival_date_hist(self, threshold: float = 0) -> pd.Timestamp:
        """
        Get the date when the portfolio balance become negative considering withdrawals on the historical data.

        The portfolio survival date (longevity date) depends on the investment strategy: asset allocation,
        rebalancing, withdrawals rate etc.

        Parameters
        ----------
        threshold : float, default 0
            The percentage of the initial investments when the portfolio balance considered voided.
            This parameter is important to use in cash flow strategies with a fixed
            withdrawal percentage (PercentageStrategy).

        Returns
        -------
        pd.Timestamp
            The portfolio survival date.

        Examples
        --------
        >>> pf = ok.Portfolio(["SPY.US", "AGG.US"], ccy="USD", first_date="2010-01", last_date="2024-10")
        >>> # set cash flow strategy
        >>> ind = ok.IndexationStrategy(pf)  # create cash flow strategy linked to the portfolio
        >>> ind.initial_investment = 10_000  # add initial investment to cash flow strategy
        >>> ind.amount = -2_500  # set annual withdrawal size
        >>> ind.frequency = "year"  # set withdrawal frequency to year
        >>> pf.dcf.cashflow_parameters = ind
        >>> # Calculate the historical survival period for the cash flow strategy
        >>> pf.dcf.survival_date_hist(threshold=0)
        Timestamp('2015-01-31 00:00:00')
        """
        ws = self.wealth_index(discounting="fv", include_negative_values=False).loc[:, self.parent.symbol]
        # TODO: change threshold to nominal value (idea)
        return helpers.Frame.get_survival_date(ws, self.discount_rate, threshold)

    @property
    def initial_investment_pv(self) -> Optional[float]:
        """
        The discounted value (PV) of the initial investments at the historical first date.

        Returns
        -------
        float or None
            The discounted value (PV) of the initial investments at the historical first date.

        Examples
        --------
        >>> # Get discounted PV value of `initial_investment` for a portfolio with 4 years of history (at 2020-04).
        >>> pf = ok.Portfolio(['EQMX.MOEX', 'SBGB.MOEX'], ccy='RUB', last_date="2024-10")
        >>> ind = ok.IndexationStrategy(pf)  # create cash flow strategy linked to the portfolio
        >>> ind.initial_investment = 10_000  # add initial investment to cash flow strategy
        >>> pf.dcf.cashflow_parameters = ind  # assign cash flow strategy to portfolio
        >>> pf.dcf.discount_rate = 0.10  # define discount rate as 10%
        >>> pf.dcf.initial_investment_pv
        6574.643143611553
        """
        if hasattr(self.cashflow_parameters, "initial_investment"):
            monlthly_discount_rate = (1 + self.discount_rate) ** (1 / settings._MONTHS_PER_YEAR) - 1
            return (
                self.cashflow_parameters.initial_investment / (1.0 + monlthly_discount_rate) ** self.parent.ror.shape[0]
            )
        else:
            return None

    @property
    def initial_investment_fv(self) -> Optional[float]:
        """
        The future value (FV) of the initial investments at the historical first date.

        When the `initial_investment` parameter is not defined, `initial_investment_fv` is set to None.

        Returns
        -------
        float or None
            The future value (FV) of the initial investments.

        Examples
        --------
        >>> # Get discounted FV of initial_investment value for a period of 10 years.
        >>> pf = ok.Portfolio(['EQMX.MOEX', 'SBGB.MOEX'], ccy='RUB')
        >>> ind = ok.IndexationStrategy(pf)  # create cash flow strategy linked to the portfolio
        >>> ind.initial_investment = 10_000  # add initial investment to cash flow strategy
        >>> pf.dcf.cashflow_parameters = ind  # assign cash flow strategy to portfolio
        >>> pf.dcf.mc.period = 10  # define forecast period
        >>> pf.dcf.discount_rate = 0.10  # define discount rate as 10%
        >>> pf.dcf.initial_investment_fv
        25937.424601000024
        """
        if hasattr(self.cashflow_parameters, "initial_investment"):
            return self.cashflow_parameters.initial_investment
        else:
            return None

    def monte_carlo_wealth(
        self, discounting: Literal["fv", "pv"] = "fv", include_negative_values: bool = True
    ) -> pd.DataFrame:
        """
        Calculate portfolio random wealth indexes with cash flows (withdrawals/contributions) using Monte Carlo simulation.

        Monte Carlo simulation generates n random monthly time series.
        Each wealth index is calculated with rate of return time series of a given distribution.

        First date of forecasted returns is portfolio last_date.
        First value for the forecasted wealth indexes is the last historical portfolio index value. It is useful
        for a chart with historical wealth index and forecasted values.

        Parameters
        ----------
        discounting : {'fv', 'pv'}, default 'fv'
            'fv' - calculate Future Value (not discounted) wealth indexes.
            'pv' - calculate Present Value (discounted) wealth indexes.
        include_negative_values : bool, default True
            If True, negative values in the wealth index are preserved.
            If False, negative values are replaced with 0.

        Returns
        -------
        DataFrame
            Table with n random wealth indexes monthly time series.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(
        ...     ['SPY.US', 'AGG.US', 'GLD.US'],
        ...     weights=[.60, .35, .05],
        ...     rebalancing_strategy=ok.Rebalance(period="month"),
        ... )
        >>> pf.dcf.set_mc_parameters(distribution="t", period=10, mc_number=100)  # Set Monte Carlo parameters
        >>> # set cash flow parameters
        >>> ind = ok.IndexationStrategy(pf)  # create cash flow strategy linked to the portfolio
        >>> ind.initial_investment = 10_000  # add initial investment to cash flow strategy
        >>> ind.amount = -500  # set withdrawal size
        >>> ind.frequency = "year"  # set withdrawal frequency
        >>> pf.dcf.cashflow_parameters = ind  # assign cash flow strategy to portfolio
        >>> pf.dcf.monte_carlo_wealth(discounting="fv").plot()
        >>> plt.legend("")  # don't show legend for each line
        >>> plt.show()
        """
        if self.cashflow_parameters is None:
            raise AttributeError("'cashflow_parameters' is not defined.")
        if self._monte_carlo_wealth_fv.empty:
            return_ts = self.mc.monte_carlo_returns_ts
            self._monte_carlo_wealth_fv = return_ts.apply(
                dcf_calculations.get_wealth_indexes_fv_with_cashflow,
                axis=0,
                args=(
                    None,  # portfolio_symbol
                    None,  # inflation_symbol
                    self.cashflow_parameters,
                    "monte_carlo",  # calculate wealth index for Monte Carlo
                ),
            )
        if not include_negative_values:
            wealth_index_fv = self._monte_carlo_wealth_fv.copy()
            wealth_index_fv = wealth_index_fv.apply(dcf_calculations.remove_negative_values, axis=0)
            # all_cells_are_nan = wealth_index_fv.isna().all(axis=1)
            # monte_carlo_wealth_fv = wealth_index_fv[~all_cells_are_nan]
            monte_carlo_wealth_fv = wealth_index_fv.fillna(0)
        else:
            monte_carlo_wealth_fv = self._monte_carlo_wealth_fv.copy()
        if discounting.lower() == "fv":
            return monte_carlo_wealth_fv
        elif discounting.lower() == "pv":
            return dcf_calculations.discount_monthly_cash_flow(monte_carlo_wealth_fv, self.discount_rate)
        else:
            raise ValueError("'discounting' must be either 'fv' or 'pv'")

    def monte_carlo_cash_flow(
        self, discounting: Literal["fv", "pv"], remove_if_wealth_index_negative: bool = True
    ) -> pd.DataFrame:
        """
        Calculate portfolio random cash flow (withdrawals/contributions) time series by Monte Carlo simulation.

        Monte Carlo simulation generates n random monthly time series of cash flows.
        Each cash flow time series corresponds to a wealth index calculated with rate of return time series of a given distribution.

        First date of forecasted returns is portfolio last_date.

        Parameters
        ----------
        discounting : {'fv', 'pv'}
            Type of discounting to apply:
            - 'fv': Future Values - nominal values without discounting
            - 'pv': Present Values - values discounted to present using the discount rate
        remove_if_wealth_index_negative : bool, default True
            If True, cash flow values are replaced with 0 when wealth index is negative (or zero).

        Returns
        -------
        DataFrame
            Table with n random cash flow monthly time series.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(
        ...     ['SPY.US', 'AGG.US', 'GLD.US'],
        ...     weights=[.60, .35, .05],
        ...     rebalancing_strategy=ok.Rebalance(period="month"),
        ... )
        >>> pf.dcf.set_mc_parameters(distribution="t", period=10, mc_number=100)  # Set Monte Carlo parameters
        >>> # set cash flow parameters
        >>> ind = ok.IndexationStrategy(pf)  # create cash flow strategy linked to the portfolio
        >>> ind.initial_investment = 10_000  # add initial investment to cash flow strategy
        >>> ind.amount = -500  # set withdrawal size
        >>> ind.frequency = "year"  # set withdrawal frequency
        >>> pf.dcf.cashflow_parameters = ind  # assign cash flow strategy to portfolio
        >>> pf.dcf.monte_carlo_cash_flow(discounting="fv").plot()
        >>> plt.legend("")  # don't show legend for each line
        >>> plt.show()
        """
        if self.cashflow_parameters is None:
            raise AttributeError("'cashflow_parameters' is not defined.")
        if self._monte_carlo_cash_flow_fv.empty:
            return_ts = self.mc.monte_carlo_returns_ts
            self._monte_carlo_cash_flow_fv = return_ts.apply(
                dcf_calculations.get_cash_flow_fv,
                axis=0,
                args=(
                    self.parent.symbol,  # portfolio_symbol
                    self.cashflow_parameters,
                    "monte_carlo",  # task
                ),
            )
        mc_cash_flow_fv = self._monte_carlo_cash_flow_fv.copy()
        if remove_if_wealth_index_negative:
            mc_wealth_index = self.monte_carlo_wealth(discounting="fv", include_negative_values=False)
            condition = mc_wealth_index == 0
            mc_cash_flow_fv[condition] = 0
        if discounting.lower() == "fv":
            return mc_cash_flow_fv
        elif discounting.lower() == "pv":
            return dcf_calculations.discount_monthly_cash_flow(mc_cash_flow_fv, self.discount_rate)
        else:
            raise ValueError("'discounting' must be either 'fv' or 'pv'")

    def plot_forecast_monte_carlo(
        self,
        backtest: bool = True,
        figsize: Optional[tuple] = None,
    ) -> Axes:
        """
        Plot Monte Carlo simulation for portfolio future wealth indexes optionally together with historical wealth index.

        Wealth index (Cumulative Wealth Index) is a time series that presents the value of portfolio over
        time period considering cash flows (portfolio withdrawals/contributions).

        Random wealth indexes are generated according to a given distribution.

        Parameters
        ----------
        backtest : bool, default True
            Include historical wealth index if True.

        figsize : tuple of (float, float), default None
            Figure size (width, height) in inches. If None, matplotlib defaults are used.

        Returns
        -------
        Axes
            Matplotlib axes object.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(
        ...     assets=["SPY.US", "AGG.US", "GLD.US"],
        ...     weights=[0.60, 0.35, 0.05],
        ...     rebalancing_strategy=ok.Rebalance(period="year"),
        ... )
        >>> # Set Monte Carlo parameters
        >>> pf.dcf.set_mc_parameters(distribution="norm", period=50, mc_number=200)
        >>> # set cash flow parameters
        >>> ind = ok.IndexationStrategy(pf)  # create cash flow strategy linked to the portfolio
        >>> ind.initial_investment = 10_000  # add initial investment to cash flow strategy
        >>> ind.amount = -500  # set withdrawal size
        >>> ind.frequency = "year"  # set withdrawal frequency
        >>> pf.dcf.cashflow_parameters = ind  # assign cash flow strategy to portfolio
        >>> pf.dcf.plot_forecast_monte_carlo(backtest=True)
        >>> plt.yscale("log")  # Y-axis has logarithmic scale
        >>> plt.show()
        """
        if backtest:
            ax = self._plot_mc_with_backtest(figsize)
        else:
            s2 = self.monte_carlo_wealth(discounting="fv", include_negative_values=False)
            ax = s2.plot(legend=None, figsize=figsize)
        self.cashflow_parameters._clear_cf_cache()
        return ax

    def _plot_mc_with_backtest(self, figsize: Optional[tuple] = None) -> Axes:
        if self.cashflow_parameters is None:
            raise AttributeError("'cashflow_parameters' is not defined.")
        original_cashflow = self.cashflow_parameters
        s1 = self.wealth_index(discounting="fv", include_negative_values=False)[self.parent.symbol]
        ax = s1.plot(legend=None, figsize=figsize)
        last_backtest_value = s1.iloc[-1]
        if last_backtest_value > 0:
            temp_cashflow = copy.copy(original_cashflow)
            try:
                self.cashflow_parameters = temp_cashflow
                self.cashflow_parameters.initial_investment = last_backtest_value
                if self.cashflow_parameters.NAME == "fixed_amount":
                    # calculate the cash flow size after backtesting period
                    months = helpers.Date.get_difference_in_months(self.parent.last_date, self.parent.first_date).n
                    years = months / settings._MONTHS_PER_YEAR
                    if self.cashflow_parameters.frequency != "none":
                        periods = years / settings.frequency_periods_per_year[self.cashflow_parameters.frequency]
                        self.cashflow_parameters.amount *= (1.0 + self.cashflow_parameters.indexation) ** periods
                s2 = self.monte_carlo_wealth(discounting="fv", include_negative_values=False)
                for s in s2:
                    s2[s].plot(legend=None, ax=ax)
            finally:
                self.cashflow_parameters = original_cashflow
        return ax

    def monte_carlo_survival_period(self, threshold: float = 0) -> pd.Series:
        """
        Generate a survival period distribution for a portfolio with cash flows by Monte Carlo simulation.

        Analyzing the result, finding "min", "max" and percentiles it's possible to see for how long
        will last the investment strategy - possible longevity period.

        Parameters
        ----------
        threshold : float, default 0
            The percentage of the initial investments when the portfolio balance considered voided.
            This parameter is important to use in cash flow strategies with a fixed
            withdrawal percentage (PercentageStrategy).

        Returns
        -------
        Series
            Survival period distribution for a portfolio with cash flows.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05])
        >>> # set Monte Carlos parameters
        >>> pf.dcf.set_mc_parameters(
        ...        distribution="t",  # use Student's distribution (t-distribution)
        ...        period=50,  # make forecast for 50 years
        ...        mc_number=200  # create 200 randow wealth indexes
        ...    )
        >>> # Set Cash Flow parameters
        >>> pc = ok.PercentageStrategy(pf)  # create PercentageStrategy linked to the portfolio
        >>> pc.initial_investment = 10_000  # add initial investments size
        >>> pc.frequency = "year"  # set cash flow frequency
        >>> pc.percentage = -0.20  # set withdrawal percentage
        >>> # Assign the strategy to Portfolio
        >>> pf.dcf.cashflow_parameters = pc
        >>> s = pf.dcf.monte_carlo_survival_period(threshold=0.10)  # the balance is considered voided at 10%
        >>> s.min()
        np.float64(10.5)
        >>> s.max()
        np.float64(33.5)
        >>> s.mean()
        np.float64(17.9055)
        >>> s.quantile(50 / 100)
        np.float64(17.5)
        """
        s2 = self.monte_carlo_wealth(discounting="fv", include_negative_values=False)
        dates: pd.Series = helpers.Frame.get_survival_date(s2, self.discount_rate, threshold)
        return dates.apply(helpers.Date.get_period_length, args=(self.parent.last_date,))

    def find_the_largest_withdrawals_size(
        self,
        goal: Literal["maintain_balance_pv", "maintain_balance_fv", "survival_period"],
        withdrawals_range: Tuple[float, float] = (0, 1),
        target_survival_period: int = 25,
        percentile: int = 20,
        threshold: float = 0,
        tolerance_rel: float = 0.10,
        iter_max: int = 20,
    ) -> Result:
        """
        Find the largest withdrawal size for Monte Carlo simulation according to a cash flow strategy.

        You can find the largest withdrawal size for three goals:

        - 'maintain_balance_pv' to keep the purchasing power of the investments after inflation
          for the whole period defined in Monte Carlo parameters.
        - 'maintain_balance_fv' to keep the nominal size of the investments for the whole period
          defined in Monte Carlo parameters.
        - 'survival_period' to keep a positive balance for a period defined by `target_survival_period`.

        The method works with IndexationStrategy and PercentageStrategy only.

        The withdrawal size defined in cash flow strategy must be negative.

        The result of finding a solution has the following parameters:
        - 'success' - whether the solution was found or not.
        - 'withdrawal_abs' - the absolute amount of withdrawal size (the best solution if found).
        - 'withdrawal_rel' - the relative amount of withdrawal size (the best solution if found).
        - 'error_rel' - characterizes how accurately the goal is fulfilled.
        - 'solutions' - the history of attempts to find solutions (withdrawal values and error level).

        The algorithm uses the bisection method to find the largest withdrawal size.

        Parameters
        ----------
        goal : {'maintain_balance_fv', 'maintain_balance_pv', 'survival_period'}
            'maintain_balance_fv' - the goal is to maintain the balance not lower than the nominal amount of the initial investment
            for the whole period defined in Monte Carlo parameters.
            'maintain_balance_pv' - the goal is to keep the purchasing power of the investments after inflation
            for the whole period defined in Monte Carlo parameters.
            'survival_period' - the goal is to keep positive balance
            for a period defined by 'target_survival_period'.

        withdrawals_range : tuple of (float, float), default (0, 1)
            The expected range of annualized withdrawal size measured as a percentage
            of the Initial Investment (CashFlow.initial_investment).
            0.01 stands for 1%. (0.02, 0.05) means that expected withdrawal is in range from 2% to 5% of Initial Investment.
            The first value is expected minimum withdrawal. The second value is expected maximum withdrawal.
            The search for a solution occurs only within this range.

        percentile : int, default 20
            The percentile of Monte-Carlo simulation distribution where the goal is achieved.
            Percentile must be from 0 to 100.
            1st or 5th percentiles are the examples of "bad" scenarios. 50th is median.
            95th or 99th are optimistic scenarios.

        threshold : float, default 0
            The percentage of initial investments when the portfolio balance is considered voided.
            Important for the "fixed_percentage" Cash flow strategy.

        target_survival_period : int, default 25
            The smallest acceptable survival period. It works with the 'survival_period' goal only.
            The value must be less than the MonteCarlo.priod parameter.

        iter_max : int, default 20
            The maximum number of iterations to find the solution.

        tolerance_rel : float, default 0.10
            The allowed tolerance for the solution. The tolerance is the largest error for the achieved goal.

        Returns
        -------
        Result
            The result of the solution search.

        Examples
        --------
        >>> pf = ok.Portfolio(
        ...     assets=["MCFTR.INDX", "RUCBTRNS.INDX"],
        ...     weights=[0.3, 0.7],
        ...     inflation=True,
        ...     ccy="RUB",
        ...     rebalancing_strategy=ok.Rebalance(period="year"),
        ... )
        >>> # Fixed Percentage strategy
        >>> pc = ok.PercentageStrategy(pf)
        >>> pc.initial_investment = 10_000
        >>> pc.frequency = "year"
        >>> # Assign a strategy
        >>> pf.dcf.cashflow_parameters = pc
        >>> # Set Monte Carlo parameters
        >>> pf.dcf.set_mc_parameters(distribution="norm", period=50, mc_number=200)
        >>> res = pf.dcf.find_the_largest_withdrawals_size(
        ...     percentile=50,
        ...     goal="survival_period",
        ...     threshold=0.05,
        ...     target_survival_period=25,
        ... )
        >>> res
        success                True
        withdrawal_abs   -917.96875
        withdrawal_rel     0.091797
        error_rel           0.00442
        attempts                 10
        dtype: object

        In the result, `withdrawal_abs` is the absolute value of the withdrawal (the first withdrawal value),
        and `withdrawal_rel` is the relative withdrawal size (the first withdrawal value divided by the initial investment).

        If the solution was not found, it is still possible to see the intermediate steps.

        >>> res.solutions
          withdrawal_abs withdrawal_rel error_rel error_rel_change
        0       -10000.0              1     0.968                0
        1        -5000.0            0.5     0.848            -0.12
        2        -2500.0           0.25    0.6082          -0.2398
        3        -1250.0          0.125   0.24816         -0.36004
        4         -625.0         0.0625   0.55576           0.3076
        5         -937.5        0.09375   0.00442         -0.55134
        """
        # Validation
        self._validate_parameters(withdrawals_range, target_survival_period, percentile, threshold, tolerance_rel)

        # Initialization
        backup_obj = self.cashflow_parameters
        backup_main_parameter = self._get_main_parameter()
        start_investment = self.cashflow_parameters.initial_investment
        expected_min_withdrawal, expected_max_withdrawal = self._get_withdrawal_bounds(
            withdrawals_range, start_investment
        )
        self._set_main_parameter(expected_max_withdrawal)

        # Bisection search
        solutions = pd.DataFrame(columns=["withdrawal_abs", "withdrawal_rel", "error_rel", "error_rel_change"])
        try:
            for iteration in range(iter_max):
                (
                    withdrawal_abs,
                    withdrawal_rel,
                    error_rel,
                    condition,
                    expected_min_withdrawal,
                    expected_max_withdrawal,
                ) = self._bisection_iteration(
                    goal,
                    percentile,
                    threshold,
                    start_investment,
                    target_survival_period,
                    expected_min_withdrawal,
                    expected_max_withdrawal,
                )

                # Record solution
                solutions.at[iteration, "withdrawal_abs"] = withdrawal_abs
                solutions.at[iteration, "withdrawal_rel"] = withdrawal_rel
                solutions.at[iteration, "error_rel"] = error_rel
                gradient = (
                    solutions.at[iteration, "error_rel"] - solutions.at[iteration - 1, "error_rel"]
                    if iteration > 0
                    else 0
                )
                solutions.at[iteration, "error_rel_change"] = gradient

                logger.info(f"Iteration {iteration}: error_rel={error_rel:.3f}, gradient={gradient:.3f}")

                # Check convergence
                if error_rel < tolerance_rel:
                    logger.info(
                        f"Solution found: {withdrawal_abs:.2f} or {withdrawal_rel * 100:.2f}% "
                        f"after {iteration + 1} steps."
                    )
                    return Result(
                        success=True,
                        withdrawal_abs=withdrawal_abs,
                        withdrawal_rel=withdrawal_rel,
                        error_rel=error_rel,
                        solutions=solutions,
                    )

            # No solution found - return best attempt
            best_idx = solutions["error_rel"].idxmin()
            best_result = solutions.loc[best_idx]
            logger.warning(
                f"Solution not found after {iter_max} steps. "
                f"Best withdrawal: {best_result['withdrawal_abs']:.2f} ({best_result['withdrawal_rel'] * 100:.2f}%) "
                f"with error: {best_result['error_rel'] * 100:.2f}%"
            )

            return Result(
                success=False,
                withdrawal_abs=best_result["withdrawal_abs"],
                withdrawal_rel=best_result["withdrawal_rel"],
                error_rel=best_result["error_rel"],
                solutions=solutions,
            )
        finally:
            self._restore_cashflow_parameters_from_backup(backup_obj, backup_main_parameter)

    def _validate_parameters(
        self,
        withdrawals_range: Tuple[float, float],
        target_survival_period: int,
        percentile: int,
        threshold: float,
        tolerance_rel: float,
    ) -> None:
        """Validate input parameters."""
        if withdrawals_range[0] > withdrawals_range[1]:
            raise ValueError("withdrawals_range[0] must be smaller than withdrawals_range[1]")
        if withdrawals_range[0] < 0 or withdrawals_range[1] > 1:
            raise ValueError("withdrawals_range[0] and withdrawals_range[1] must be in range from 0 to 1.")
        if target_survival_period > self.mc.period * (1 - tolerance_rel):
            # The Monte Carlo period length must be greater than target_survival_period by at least the tolerance_rel.
            # Otherwise, solutions with very low withdrawal rates will be successful,
            # since for them, survival_period will be equal to the Monte Carlo period,
            # and the error will be lower than the tolerance_rel.
            raise ValueError(
                f"target_survival_period must be less than Monte Carlo simulation period ({self.mc.period})."
            )
        if not 0 <= percentile <= 100:
            raise ValueError("percentile must be between 0 and 100")
        if not 0 <= threshold <= 1:
            raise ValueError("threshold must be between 0 and 1")

    def _update_parameter(self, delta: float, increase: bool) -> None:
        """Update withdrawal parameter using bisection step."""
        sign = -1 if increase else 1
        if isinstance(self.cashflow_parameters, cf.IndexationStrategy):
            self.cashflow_parameters.amount += sign * delta / 2
        elif isinstance(self.cashflow_parameters, cf.PercentageStrategy):
            self.cashflow_parameters.percentage += sign * delta / 2
        else:
            raise ValueError(
                "This method works with IndexationStrategy, PercentageStrategy cash flow strategies and their subclasses only."
            )

    def _calculate_goal_metrics(
        self, goal: str, percentile: int, threshold: float, start_investment: float, target_survival_period: int
    ) -> Tuple[bool, float]:
        """
        Calculate whether goal condition is met and relative error.

        Returns
        -------
        Tuple[bool, float]
            (condition_met, error_rel)
        """
        sp_at_quantile = self.monte_carlo_survival_period(threshold=threshold).quantile(percentile / 100)

        if goal in [WithdrawalGoal.MAINTAIN_BALANCE_FV, WithdrawalGoal.MAINTAIN_BALANCE_PV]:
            discounting = "pv" if goal == WithdrawalGoal.MAINTAIN_BALANCE_PV else "fv"
            s = self.monte_carlo_wealth(discounting, True)
            wealth_at_quantile = s.iloc[-1, :].quantile(percentile / 100)
            condition = (wealth_at_quantile >= start_investment) and (sp_at_quantile == self.mc.period)
            error_rel = abs(wealth_at_quantile - start_investment) / start_investment
            logger.debug(
                f"Goal: {goal}, wealth_at_quantile={wealth_at_quantile:.2f}, main_parameter={self._get_main_parameter():.3f}"
            )
            return condition, error_rel
        elif goal == WithdrawalGoal.SURVIVAL_PERIOD:
            condition = sp_at_quantile >= target_survival_period
            error_rel = abs(sp_at_quantile - target_survival_period) / target_survival_period
            logger.debug(
                f"Goal: {goal}, sp_at_quantile={sp_at_quantile:.2f}, main_parameter={self._get_main_parameter():.3f}"
            )
            return condition, error_rel
        else:
            raise ValueError("The goal can be: maintain_balance_fv, maintain_balance_pv or survival_period.")

    def _calculate_withdrawal_metrics(self, main_parameter: float, start_investment: float) -> Tuple[float, float]:
        """Calculate absolute and relative withdrawal values.

        Returns
        -------
        Tuple[float, float]
            (withdrawal_abs, withdrawal_rel)
        """
        if isinstance(self.cashflow_parameters, cf.IndexationStrategy):
            withdrawal_abs = main_parameter
            withdrawal_rel = abs(main_parameter / start_investment * self.cashflow_parameters.periods_per_year)
        elif isinstance(self.cashflow_parameters, cf.PercentageStrategy):
            withdrawal_abs = main_parameter * start_investment / self.cashflow_parameters.periods_per_year
            withdrawal_rel = abs(main_parameter)
        else:
            raise ValueError(
                "This method works with IndexationStrategy, PercentageStrategy cash flow strategies and their subclasses only."
            )

        return withdrawal_abs, withdrawal_rel

    def _bisection_iteration(
        self,
        goal: str,
        percentile: int,
        threshold: float,
        start_investment: float,
        target_survival_period: int,
        expected_min_withdrawal: float,
        expected_max_withdrawal: float,
    ) -> Tuple[float, float, float, bool, float, float]:
        """Perform one iteration of bisection search.

        Returns
        -------
        Tuple containing: withdrawal_abs, withdrawal_rel, error_rel, condition,
                          new_min_withdrawal, new_max_withdrawal
        """
        main_parameter = self._get_main_parameter()
        condition, error_rel = self._calculate_goal_metrics(
            goal, percentile, threshold, start_investment, target_survival_period
        )
        withdrawal_abs, withdrawal_rel = self._calculate_withdrawal_metrics(main_parameter, start_investment)

        # Update bounds based on condition
        if condition:
            expected_min_withdrawal = main_parameter
            delta = abs(expected_max_withdrawal - main_parameter)
            self._update_parameter(delta, increase=True)
            logger.debug("Increasing withdrawal")
        else:
            expected_max_withdrawal = main_parameter
            delta = abs(main_parameter - expected_min_withdrawal)
            self._update_parameter(delta, increase=False)
            logger.debug("Decreasing withdrawal")

        return (withdrawal_abs, withdrawal_rel, error_rel, condition, expected_min_withdrawal, expected_max_withdrawal)

    def _restore_cashflow_parameters_from_backup(
        self, backup_obj: cf.CashFlow | None, backup_main_parameter: float | None = None
    ) -> None:
        self.cashflow_parameters = backup_obj
        if self.cashflow_parameters is None:
            return

        if backup_main_parameter is not None:
            self._set_main_parameter(backup_main_parameter)
        else:
            self.cashflow_parameters._clear_cf_cache()

    def _get_withdrawal_bounds(
        self, withdrawals_range: Tuple[float, float], start_investment: float
    ) -> Tuple[float, float]:
        """Calculate min and max withdrawal bounds based on strategy type.

        Returns
        -------
        Tuple[float, float]
            (expected_min_withdrawal, expected_max_withdrawal)
        """
        if isinstance(self.cashflow_parameters, cf.IndexationStrategy):
            return (
                -withdrawals_range[0] * start_investment / self.cashflow_parameters.periods_per_year,
                -withdrawals_range[1] * start_investment / self.cashflow_parameters.periods_per_year,
            )
        elif isinstance(self.cashflow_parameters, cf.PercentageStrategy):
            return -withdrawals_range[0], -withdrawals_range[1]
        else:
            raise ValueError(
                "This method works with IndexationStrategy, PercentageStrategy cash flow strategies and their subclasses only."
            )

    def _get_main_parameter(self) -> float:
        """Get the main withdrawal parameter based on strategy type."""
        if isinstance(self.cashflow_parameters, cf.IndexationStrategy):
            return self.cashflow_parameters.amount
        elif isinstance(self.cashflow_parameters, cf.PercentageStrategy):
            return self.cashflow_parameters.percentage
        else:
            raise ValueError(
                "This method works with IndexationStrategy, PercentageStrategy cash flow strategies and their subclasses only."
            )

    def _set_main_parameter(self, value: float) -> None:
        """Set the main withdrawal parameter based on strategy type."""
        if isinstance(self.cashflow_parameters, cf.IndexationStrategy):
            self.cashflow_parameters.amount = value
        elif isinstance(self.cashflow_parameters, cf.PercentageStrategy):
            self.cashflow_parameters.percentage = value
        else:
            raise ValueError(
                "This method works with IndexationStrategy, PercentageStrategy cash flow strategies and their subclasses only."
            )



import time
import itertools
from typing import List, Tuple, Dict, Optional

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
from matplotlib.axes import Axes

from scipy.optimize import minimize

from okama import asset_list, settings
from okama.common.helpers import helpers
from okama.common.helpers.rebalancing import Rebalance

import logging

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


class EfficientFrontier(asset_list.AssetList):
    """
    Efficient Frontier with multi-period optimization.

    In multi-period optimization, portfolios are rebalanced according to a rebalancing strategy.

    Rebalancing is the process by which an investor restores a portfolio to its target allocation
    by selling and buying assets. After rebalancing, the portfolio assets have their target weights.

    Parameters
    ----------
    assets : list[str]
        List of assets. Could include tickers or asset-like objects (`Asset`, `Portfolio`).
        Must contain at least two assets.

    first_date : str, default None
        First date of monthly return time series.
        If None the first date is calculated automatically as the oldest available date for the listed assets.

    last_date : str, default None
        Last date of monthly return time series.
        If None the last date is calculated automatically as the newest available date for the listed assets.

    ccy : str, default 'USD'
        Base currency for the list of assets. All risk metrics and returns are adjusted to the base currency.

    bounds : tuple[tuple[float, float], ...] or None, default None
        Bounds for asset weights. Each asset weight is constrained within the corresponding (min, max) pair.
        For example, `((0.0, 0.5), (0.0, 1.0))` restricts the first asset to [0%, 50%] and leaves the second
        asset unconstrained.

    inflation : bool, default False
        Defines whether to take inflation data into account in the calculations.
        Including inflation could limit available data (`first_date`, `last_date`) as the inflation data is
        usually published with a one-month delay.

    rebalancing_strategy : Rebalance, default Rebalance(period='year')
        Rebalancing strategy used to generate portfolio return series during optimization.
        Only `rebalancing_strategy.period` is used; `abs_deviation` and `rel_deviation` are ignored.

    n_points : int, default 20
        Number of points in the Efficient Frontier.

    full_frontier : bool, default True
        Defines whether to show the full Efficient Frontier or only its upper part.
        If 'False' Efficient Frontier has only the points with the return above Global Minimum Volatility (GMV) point.

    verbose : bool, default False
        If verbose=True calculates elapsed time for each point and the total elapsed time.

    ticker_names : bool, default True
        Defines whether to include full names of assets in the optimization report or only tickers.

    Notes
    -----
    For classic single-period (monthly rebalanced, constant-weight) optimization, use `EfficientFrontierSingle`.
    """

    _FTOL = (1e-06, 1e-05, 1e-3, 1e-02)  # possible tolerance values for the optimizer

    def __init__(
        self,
        assets: Optional[List[str]] = None,
        *,
        first_date: Optional[str] = None,
        last_date: Optional[str] = None,
        ccy: str = "USD",
        bounds: Optional[Tuple[Tuple[float, ...], ...]] = None,
        inflation: bool = False,
        full_frontier: bool = True,
        rebalancing_strategy: Rebalance = Rebalance(period="year"),
        n_points: int = 20,
        verbose: bool = False,
        ticker_names: bool = True,
    ):
        if len(assets) < 2:
            raise ValueError("The number of symbols cannot be less than two")
        super().__init__(
            assets=assets,
            first_date=first_date,
            last_date=last_date,
            ccy=ccy,
            inflation=inflation,
        )

        self._bounds = None
        self.bounds = bounds
        self.rebalancing_strategy = rebalancing_strategy
        self.n_points = n_points
        self.ticker_names = ticker_names
        self.verbose = verbose
        self.full_frontier = full_frontier
        self._ef_points = pd.DataFrame(dtype=float)
        self._mdp_points = pd.DataFrame(dtype=float)
        self._ror_cache: Dict[tuple, pd.Series] = {}  # Cache for portfolio return time series by weights

    def __repr__(self):
        dic = {
            "symbols": self.symbols,
            "currency": self._currency.ticker,
            "first_date": self.first_date.strftime("%Y-%m"),
            "last_date": self.last_date.strftime("%Y-%m"),
            "period_length": self._pl_txt,
            "rebalancing_period": self.rebalancing_strategy.period,
            "rebalancing_abs_deviation": self.rebalancing_strategy.abs_deviation,
            "rebalancing_rel_deviation": self.rebalancing_strategy.rel_deviation,
            "bounds": self.bounds,
            "inflation": self.inflation if hasattr(self, "inflation") else "None",
            "n_points": self.n_points,
        }
        return repr(pd.Series(dic))

    @property
    def bounds(self) -> Tuple[Tuple[float, ...], ...]:
        """
        Return bounds for the assets weights.

        Bounds are used in optimization. Each asset can have weights limitation from 0 to 1.0.

        If an asset has limitation for 10 to 20% bounds are defined as (0.1, 0.2).
        bounds = ((0, .5), (0, 1)) shows that in Portfolio with two assets first one has weight limitations
        from 0 to 50%. The second asset has no limitations.

        Returns
        -------
        tuple of ((float, float),...)
            Weights bounds used for portfolio optimization.

        Examples
        --------
        >>> two_assets = ok.EfficientFrontierSingle(['SPY.US', 'AGG.US'])
        >>> two_assets.bounds
        ((0.0, 1.0), (0.0, 1.0))

        By default there are no limitations for assets weights.
        Bounds can be set for a Efficient Frontier object.

        >>> two_assets.bounds = ((0.5, 0.9), (0, 1.0))

        Now the optimization is bounded (SPY has weights limits from 50 to 90%).
        """
        return self._bounds

    @bounds.setter
    def bounds(self, bounds):

        self._ef_points = pd.DataFrame(dtype=float)
        self._mdp_points = pd.DataFrame(dtype=float)

        if bounds:
            if len(bounds) != len(self.symbols):
                raise ValueError(
                    f"The number of symbols ({len(self.symbols)}) "
                    f"and the length of bounds ({len(bounds)}) should be equal."
                )
            self._bounds = bounds
        else:
            self._bounds = ((0.0, 1.0),) * len(self.symbols)  # an N-tuple of 2-tuples

    @property
    def n_points(self) -> int:
        """
        Return or set number of points in the Efficient Frontier.

        Returns
        -------
        int
            Number of points in the Efficient Frontier.

        Examples
        --------
        >>> frontier = ok.EfficientFrontier(['SPY.US', 'BND.US'])
        >>> frontier.n_points  # default number of points
        20
        """
        return self._n_points

    @n_points.setter
    def n_points(self, n_points: int):
        if not isinstance(n_points, int):
            raise ValueError("n_points should be an integer")
        if n_points <= 0:
            raise ValueError("n_points should be greater than zero")
        self._clear_cache()
        self._n_points = n_points

    def _clear_cache(self):
        self._ef_points = pd.DataFrame(dtype=float)  # renew EF points DataFrame
        self._mdp_points = pd.DataFrame(dtype=float)  # renew MDP points DataFrame
        self._ror_cache = {}  # clear portfolio return time series cache

    def _get_portfolio_ror_ts(self, weights: np.ndarray) -> pd.Series:
        """
        Get portfolio return time series with caching based on weights.

        This method caches the results of portfolio return calculations to avoid
        redundant computations during optimization when the same weights are evaluated
        multiple times (e.g., in objective function and constraints).

        Parameters
        ----------
        weights : np.ndarray
            Portfolio weights for each asset.

        Returns
        -------
        pd.Series
            Monthly rate of return time series for the rebalanced portfolio.
        """
        # Round weights to avoid floating point precision issues in cache keys
        weights_key = tuple(np.round(weights, decimals=10))

        if weights_key not in self._ror_cache:
            rebalance = Rebalance(period=self.rebalancing_strategy.period)
            self._ror_cache[weights_key] = rebalance.return_ror_ts_ef(weights, self.assets_ror)

        return self._ror_cache[weights_key]

    @property
    def rebalancing_strategy(self) -> Rebalance:
        """
        Rebalancing strategy used to compute portfolio return series during optimization.

        Only `rebalancing_strategy.period` is used in the optimization; `abs_deviation` and `rel_deviation`
        are ignored.

        Returns
        -------
        Rebalance
            Rebalancing strategy.

        Examples
        --------
        >>> frontier = ok.EfficientFrontier(['SPY.US', 'BND.US'])
        >>> frontier.rebalancing_strategy.period
        'year'

        >>> frontier.rebalancing_strategy = ok.Rebalance(period='none')
        """
        return self._rebalancing_strategy

    @rebalancing_strategy.setter
    def rebalancing_strategy(self, rebalancing_strategy: Rebalance):
        if isinstance(rebalancing_strategy, Rebalance):
            if rebalancing_strategy.abs_deviation is not None or rebalancing_strategy.rel_deviation is not None:
                logger.warning(
                    "Absolute and relative constraints (abs_deviation, rel_deviation) are not considered "
                    "during portfolio optimization. Only the rebalancing period is used."
                )
            self._clear_cache()
            self._rebalancing_strategy = rebalancing_strategy
        else:
            raise ValueError("rebalancing_strategy must be of type Rebalance")

    @property
    def ticker_names(self):
        """
        Return or set whether to show tickers or full stock names in the reports.

        Returns
        -------
        bool
            True - for tickers.
            False - for full stock (index) names.
        """
        return self._tickers

    @ticker_names.setter
    def ticker_names(self, tickers: bool):
        if not isinstance(tickers, bool):
            raise ValueError("tickers should be True or False")
        self._tickers = tickers

    @property
    def verbose(self):
        """
        Return or set whether to show technical information during the optimization.

        Returns
        -------
        bool
        """
        return self._verbose

    @verbose.setter
    def verbose(self, verbose: bool):
        if not isinstance(verbose, bool):
            raise ValueError("verbose should be True or False")
        self._clear_cache()
        self._verbose = verbose

    def get_most_diversified_portfolio(
        self,
        target_return: Optional[float] = None,
    ) -> Dict[str, float]:
        """
        Calculate assets weights and portfolio metrics for the most diversified portfolio within bounds.

        The most diversified portfolio is defined as the portfolio with the maximum Diversification Ratio.

        Parameters
        ----------
        target_return : float, default None
            Target Compound Annual Growth Rate (CAGR) for the portfolio. If provided, the optimizer searches for a
            portfolio with the target CAGR and the maximum Diversification Ratio. If None, the global most diversified
            portfolio is returned.

        Returns
        -------
        dict[str, float]
            Mapping with asset weights (keys are tickers or asset names depending on `ticker_names`) and portfolio
            metrics: 'CAGR', 'Risk', and 'Diversification ratio'.

        Examples
        --------
        >>> ls4 = ['SPY.US', 'AGG.US', 'VNQ.US', 'GLD.US']
        >>> x = ok.EfficientFrontier(assets=ls4, ccy='USD', last_date='2021-12')
        >>> x.get_most_diversified_portfolio()  # get a global most diversified portfolio
        {'SPY.US': 0.19612726258395477,
        'AGG.US': 0.649730553241489,
        'VNQ.US': 0.020096313783052246,
        'GLD.US': 0.13404587039150392,
        'CAGR': 0.062355715886719176,
        'Risk': 0.05510135025563423,
        'Diversification ratio': 1.5665720501693001}

        It is possible to get the most diversified portfolio for a given target CAGR.

        >>> x.get_most_diversified_portfolio(target_return=0.10)
        {'SPY.US': 0.3389762570274293,
        'AGG.US': 0.12915657041748244,
        'VNQ.US': 0.15083042115027034,
        'GLD.US': 0.3810367514048179,
        'CAGR': 0.09370688842211439,
        'Risk': 0.11725067815643951,
        'Diversification ratio': 1.4419864802150442}
        """
        ror = self.assets_ror
        n = self.assets_ror.shape[1]
        init_guess = np.repeat(1 / n, n)

        def objective_function(w):
            # Diversification Ratio
            assets_risk = ror.std()
            assets_mean_return = self.assets_ror.mean()
            assets_annualized_risk = helpers.Float.annualize_risk(assets_risk, assets_mean_return)
            weights = np.asarray(w)
            assets_sigma_weighted_sum = weights.T @ assets_annualized_risk

            portfolio_ror = self._get_portfolio_ror_ts(w)
            portfolio_mean_return_monthly = portfolio_ror.mean()
            portfolio_risk_monthly = portfolio_ror.std()

            objective_function.annual_risk = helpers.Float.annualize_risk(
                portfolio_risk_monthly, portfolio_mean_return_monthly
            )
            return -assets_sigma_weighted_sum / objective_function.annual_risk

        # construct the constraints
        weights_sum_to_1 = {"type": "eq", "fun": lambda weights: np.sum(weights) - 1}
        cagr_is_target = {
            "type": "eq",
            "fun": lambda weights: target_return - self._get_cagr(weights),
        }
        constraints = (weights_sum_to_1,) if target_return is None else (weights_sum_to_1, cagr_is_target)

        # set optimizer
        weights = minimize(
            objective_function,
            init_guess,
            method="SLSQP",
            options={"disp": False},
            constraints=constraints,
            bounds=self.bounds,
        )
        if weights.success:
            # CAGR calculation
            cagr = self._get_cagr(weights.x)
            asset_labels = self.symbols if self.ticker_names else list(self.names.values())
            point = {x: y for x, y in zip(asset_labels, weights.x)}
            point["CAGR"] = cagr
            point["Risk"] = objective_function.annual_risk
            point["Diversification ratio"] = -weights.fun
            return point
        else:
            raise RecursionError("No solutions where found")

    def get_tangency_portfolio(self, rf_return: float = 0, rate_of_return: str = "cagr") -> dict:
        """
        Calculate asset weights, risk and return values for tangency portfolio within given bounds.

        Tangency portfolio or Maximum Sharpe Ratio (MSR) is the point on the Efficient Frontier where
        Sharpe Ratio reaches its maximum.

        The Sharpe ratio is the average annual return in excess of the risk-free rate
        per unit of risk (annualized standard deviation).

        Bounds are defined with 'bounds' property.

        Parameters
        ----------
        rf_return : float, default 0
            Risk-free rate of return.

        rate_of_return : {'cagr', 'mean_return'}, default 'cagr'
            Return definition used to calculate Sharpe ratio.

            - 'cagr': Compound Annual Growth Rate.
            - 'mean_return': Arithmetic mean return (annualized).

        Returns
        -------
        dict
             Weights of assets, risk and return of the tangency portfolio.

        Examples
        --------
        >>> three_assets = ['SPY.US', 'AGG.US', 'GLD.US']
        >>> ef = ok.EfficientFrontier(assets=three_assets, ccy='USD', last_date='2022-06')
        >>> msr = ef.get_tangency_portfolio(rf_return=0.03)  # risk free rate of return is 3%
        >>> sorted(msr)
        ['Rate_of_return', 'Risk', 'Weights']

        To calculate tangency portfolio parameters for arithmetic mean set rate_of_return='mean_return':

        >>> msr_mean = ef.get_tangency_portfolio(rate_of_return="mean_return", rf_return=0.03)
        >>> sorted(msr_mean)
        ['Rate_of_return', 'Risk', 'Weights']
        """
        n = self.assets_ror.shape[1]
        init_guess = np.repeat(1 / n, n)

        def of_arithmetic_mean(w):
            # Sharpe ratio with arithmetic mean
            portfolio_ror = self._get_portfolio_ror_ts(w)
            mean_return_monthly = portfolio_ror.mean()
            risk_monthly = portfolio_ror.std()
            of_arithmetic_mean.rate_of_return = helpers.Float.annualize_return(mean_return_monthly)
            of_arithmetic_mean.risk = helpers.Float.annualize_risk(risk_monthly, mean_return_monthly)
            return -(of_arithmetic_mean.rate_of_return - rf_return) / of_arithmetic_mean.risk

        def of_geometric_mean(w):
            # Sharpe ratio with CAGR
            portfolio_ror = self._get_portfolio_ror_ts(w)
            mean_return_monthly = portfolio_ror.mean()
            of_geometric_mean.rate_of_return = helpers.Frame.get_cagr(portfolio_ror)
            # Risk
            risk_monthly = portfolio_ror.std()
            of_geometric_mean.risk = helpers.Float.annualize_risk(risk_monthly, mean_return_monthly)
            return -(of_geometric_mean.rate_of_return - rf_return) / of_geometric_mean.risk

        if rate_of_return.lower() in {"cagr", "mean_return"}:
            rate_of_return = rate_of_return.lower()
        else:
            raise ValueError("rate_of_return must be 'cagr' or 'mean_return'")

        objective_function = of_geometric_mean if rate_of_return == "cagr" else of_arithmetic_mean
        # construct the constraints
        weights_sum_to_1 = {"type": "eq", "fun": lambda weights: np.sum(weights) - 1}
        weights = minimize(
            objective_function,
            init_guess,
            method="SLSQP",
            options={"disp": False},
            constraints=(weights_sum_to_1,),
            bounds=self.bounds,
        )
        if weights.success:
            return {
                "Weights": weights.x,
                "Rate_of_return": objective_function.rate_of_return,
                "Risk": objective_function.risk,
            }
        else:
            raise RecursionError("No solutions where found")

    @property
    def gmv_monthly_weights(self) -> np.ndarray:
        """
        Calculate asset weights of the Global Minimum Volatility (GMV) portfolio. The objective function is
        monthly risk (standard deviation of return).

        Global Minimum Volatility portfolio is a portfolio with the lowest risk of all possible.
        Along the Efficient Frontier, the left-most point is a portfolio with minimum risk when compared to
        all possible portfolios of risky assets.

        Returns
        -------
        numpy.ndarray
            GMV portfolio assets weights.

        Examples
        --------
        >>> frontier = ok.EfficientFrontier(['SPY.US', 'AGG.US'])
        >>> frontier.gmv_monthly_weights
        array([0.0578446, 0.9421554])
        """
        n = self.assets_ror.shape[1]
        init_guess = np.repeat(1 / n, n)

        # Set the objective function
        def objective_function(w):
            risk = self._get_portfolio_ror_ts(w).std()
            return risk

        # construct the constraints
        weights_sum_to_1 = {"type": "eq", "fun": lambda weights: np.sum(weights) - 1}
        weights = minimize(
            objective_function,
            init_guess,
            method="SLSQP",
            options={"disp": False},
            constraints=(weights_sum_to_1,),
            bounds=self.bounds,
        )
        return weights.x

    @property
    def gmv_annual_weights(self) -> np.ndarray:
        """
        Calculate asset weights of the Global Minimum Volatility (GMV) portfolio. The objective function is
        annualized risk (standard deviation of return).

        Global Minimum Volatility portfolio is a portfolio with the lowest risk of all possible.
        Along the Efficient Frontier, the left-most point is a portfolio with minimum risk when compared to
        all possible portfolios of risky assets.

        Returns
        -------
        numpy.ndarray
            GMV portfolio assets weights.

        Examples
        --------
        >>> frontier = ok.EfficientFrontier(['SPY.US', 'AGG.US'])
        >>> frontier.gmv_monthly_weights
        array([0.05373824, 0.94626176])
        """
        n = self.assets_ror.shape[1]
        init_guess = np.repeat(1 / n, n)

        # Set the objective function
        def objective_function(w):
            ts = self._get_portfolio_ror_ts(w)
            mean_return = ts.mean()
            risk = ts.std()
            return helpers.Float.annualize_risk(risk=risk, mean_return=mean_return)

        # construct the constraints
        weights_sum_to_1 = {"type": "eq", "fun": lambda weights: np.sum(weights) - 1}
        weights = minimize(
            objective_function,
            init_guess,
            method="SLSQP",
            options={"disp": False},
            constraints=(weights_sum_to_1,),
            bounds=self.bounds,
        )
        return weights.x

    def _get_gmv_monthly(self) -> Tuple[float, float]:
        """
        Calculate the risk and return (mean, monthly) of the Global Minimum Volatility portfolio.

        Global Minimum Volatility portfolio is a portfolio with the lowest risk of all possible.
        """
        ts = self._get_portfolio_ror_ts(self.gmv_monthly_weights)
        return ts.std(), ts.mean()

    @property
    def gmv_annual_values(self) -> Tuple[float, float]:
        """
        Calculate the annualized risk (standard deviation) and CAGR of the Global Minimum Volatility portfolio.

        Global Minimum Volatility portfolio is a portfolio with the lowest risk of all possible.
        Compound annual growth rate (CAGR) is the rate of return that would be required for an investment to grow from
        its initial to its final value, assuming all incomes were reinvested.

        Returns
        -------
        tuple
            Annualized value of risk (standard deviation),
            Compound annual growth rate (CAGR)
            for Global Minimum Volatility portfolio (GMV).

        Examples
        --------
        >>> frontier = ok.EfficientFrontier(['SPY.US', 'AGG.US'])
        >>> frontier.gmv_annual_values
        (0.03695845106087943, 0.04418318557516887)
        """
        returns = self._get_portfolio_ror_ts(self.gmv_annual_weights)
        return (
            helpers.Float.annualize_risk(returns.std(), returns.mean()),
            (returns + 1.0).prod() ** (settings._MONTHS_PER_YEAR / returns.shape[0]) - 1.0,
        )

    @property
    def global_max_return_portfolio(self) -> dict:
        """
        Find a portfolio with global max CAGR.

        Compound annual growth rate (CAGR) is the rate of return that would be required for an investment to grow from
        its initial to its final value, assuming all incomes were reinvested.

        The objective function is Accumulated return for rebalanced portfolio time series for the period
        from 'first_date' to 'last_date'.

        Returns
        -------
        dict
            Weights of assets, CAGR, annualized risk, monthly risk.

        Examples
        --------
        >>> frontier = ok.EfficientFrontier(['SPY.US', 'AGG.US'])
        >>> frontier.global_max_return_portfolio
        {'Weights': array([1., 0.]), 'CAGR': 0.10797159166196812, 'Risk': 0.1583011735798155, 'Risk_monthly': 0.0410282468594492}
        """
        n = self.assets_ror.shape[1]  # Number of assets
        init_guess = np.repeat(1 / n, n)

        # Set the objective function
        def objective_function(w):
            # Accumulated return for rebalanced portfolio time series
            objective_function.returns = self._get_portfolio_ror_ts(w)
            accumulated_return = (objective_function.returns + 1.0).prod() - 1.0
            return -accumulated_return

        # construct the constraints
        weights_sum_to_1 = {"type": "eq", "fun": lambda weights: np.sum(weights) - 1}
        weights = minimize(
            objective_function,
            init_guess,
            method="SLSQP",
            options={
                "disp": False,
                "maxiter": 100,
                "ftol": self._FTOL[0],
            },
            constraints=(weights_sum_to_1,),
            bounds=self.bounds,
        )
        portfolio_ts = objective_function.returns
        mean_return = portfolio_ts.mean()
        portfolio_risk = portfolio_ts.std()
        point = {
            "Weights": weights.x,
            "CAGR": (1 - weights.fun) ** (settings._MONTHS_PER_YEAR / self.assets_ror.shape[0]) - 1,
            "Risk": helpers.Float.annualize_risk(portfolio_risk, mean_return),
            "Risk_monthly": portfolio_risk,
        }
        return point

    def _get_cagr(self, weights: np.ndarray) -> float:
        ts = self._get_portfolio_ror_ts(weights)
        acc_return = (ts + 1.0).prod() - 1.0
        return (1.0 + acc_return) ** (settings._MONTHS_PER_YEAR / ts.shape[0]) - 1.0

    def minimize_risk(self, target_value: float) -> Dict[str, float]:
        """
        Calculate the portfolio properties to minimize annualized risk at the target CAGR.

        This method finds the portfolio weights that minimize the annualized risk (standard deviation)
        while achieving a specified target Compound Annual Growth Rate (CAGR).

        The optimization is performed for a rebalanced portfolio over the period from 'first_date' to 'last_date'.
        CAGR is the rate of return required for an investment to grow from its initial to its final value,
        assuming all incomes were reinvested.

        Parameters
        ----------
        target_value : float
            Target Compound Annual Growth Rate (CAGR) for the portfolio.
            Should be a decimal value (e.g., 0.107 for 10.7% annual return).

        Returns
        -------
        dict
            Dictionary containing:
            - Asset weights (one key per asset symbol or name)
            - 'CAGR': Target CAGR value
            - 'Risk': Minimized annualized risk (standard deviation)
            - 'Weights': Array of optimal weights
            - 'iterations': Number of optimization iterations performed

        Raises
        ------
        RecursionError
            If no solution is found for the given target CAGR value.

        Examples
        --------
        >>> frontier = ok.EfficientFrontier(['SPY.US', 'AGG.US'])
        >>> point = frontier.minimize_risk(0.08)
        >>> round(point["CAGR"], 2)
        0.08
        """

        n = self.assets_ror.shape[1]  # number of assets
        init_guess = np.repeat(1 / n, n)  # initial weights

        max_ratio_data = self._max_ratio_asset_right_to_max_cagr

        if max_ratio_data is not None:
            init_guess = np.repeat(0, n)  # clear weights
            init_guess[self._min_ratio_asset["list_position"]] = 1.0

        def objective_function(w):
            # annual risk
            ts = self._get_portfolio_ror_ts(w)
            risk_monthly = ts.std()
            objective_function.mean_return = ts.mean()
            return helpers.Float.annualize_risk(risk_monthly, objective_function.mean_return)

        # construct the constraints
        weights_sum_to_1 = {"type": "eq", "fun": lambda weights: np.sum(weights) - 1}
        cagr_is_target = {
            "type": "eq",
            "fun": lambda weights: target_value - self._get_cagr(weights),
        }
        # for i in range(4):
        weights = minimize(
            objective_function,
            init_guess,
            method="SLSQP",
            options={
                "disp": False,
                "maxiter": 80,
                "ftol": self._FTOL[0],
            },
            constraints=(weights_sum_to_1, cagr_is_target),
            bounds=self.bounds,
        )

        # Calculate points of EF given optimal weights
        if weights.success:
            asset_labels = self.symbols if self.ticker_names else list(self.names.values())
            point = dict(zip(asset_labels, weights.x))
            point["CAGR"] = target_value
            point["Mean return"] = objective_function.mean_return * settings._MONTHS_PER_YEAR
            point["Risk"] = weights.fun
            point["Weights"] = weights.x
            point["iterations"] = weights.nit
            # break
        if not weights.success:
            raise RecursionError(f"No solution found for target CAGR value: {target_value}.")
        return point

    def _maximize_risk(self, target_return: float) -> Dict[str, float]:
        """
        Calculate the portfolio properties to maximize annualized value of risk at the target CAGR.

        The objective function is Annualized risk (standard deviation) for rebalanced portfolio time series
        for the period from 'first_date' to 'last_date'.

        Parameters
        ----------
        target_return : float
            Target CAGR value

        Returns
        -------
        dict
            Weights of assets, CAGR, annualized risk and technical optimization parameters.
        """
        n = self.assets_ror.shape[1]  # number of assets

        def objective_function(w):
            # annual risk
            ts = self._get_portfolio_ror_ts(w)
            risk_monthly = ts.std()
            objective_function.mean_return = ts.mean()
            result = -helpers.Float.annualize_risk(risk_monthly, objective_function.mean_return)
            return result

        # construct the constraints
        weights_sum_to_1 = {"type": "eq", "fun": lambda weights: np.sum(weights) - 1}
        cagr_is_target = {
            "type": "eq",
            "fun": lambda weights: target_return - self._get_cagr(weights),
        }

        constraints = [weights_sum_to_1, cagr_is_target]

        # Prepare initial guess hints
        init_guesses = dict()

        # Hint 1: max ratio asset
        if self._max_ratio_asset_right_to_max_cagr:
            init_guess_1 = np.repeat(0, n)
            init_guess_1[self._max_ratio_asset_right_to_max_cagr["list_position"]] = 1.0
            init_guesses["max_ratio_asset_right_to_max_cagr"] = init_guess_1

        # Hint 2: Global max return portfolio
        if hasattr(self, "_global_max_return_portfolio_weights"):
            init_guesses["global_max_return_portfolio"] = self._global_max_return_portfolio_weights.copy()
        else:
            global_max = self.global_max_return_portfolio
            init_guesses["global_max_return_portfolio"] = global_max["Weights"].copy()
            self._global_max_return_portfolio_weights = global_max["Weights"].copy()  # caching result

        solution = None
        for init_guess_key, init_guess_value in init_guesses.items():
            weights = minimize(
                objective_function,
                init_guess_value,
                method="SLSQP",
                options={
                    "disp": False,
                    "ftol": self._FTOL[0],
                    "maxiter": 80,
                },
                constraints=constraints,
                bounds=self.bounds,
            )

            # Calculate points of EF given optimal weights
            if weights.success:
                asset_labels = self.symbols if self.ticker_names else list(self.names.values())
                solution = dict(zip(asset_labels, weights.x))
                solution["CAGR"] = target_return
                solution["Mean return"] = objective_function.mean_return * settings._MONTHS_PER_YEAR
                solution["Risk"] = -weights.fun
                solution["Weights"] = weights.x
                solution["iterations"] = weights.nit
                solution["init_guess"] = init_guess_key
                break

        if solution is None:
            raise RecursionError(f"No solution found for target CAGR value: {target_return}.")
        return solution

    @property
    def _max_cagr_asset(self) -> dict:
        """
        Find an asset with max CAGR.
        """
        max_asset_cagr = helpers.Frame.get_cagr(self.assets_ror).max()
        ticker_with_largest_cagr = helpers.Frame.get_cagr(self.assets_ror).nlargest(1, keep="first").index.values[0]
        return {
            "max_asset_cagr": max_asset_cagr,
            "ticker_with_largest_cagr": ticker_with_largest_cagr,
            "list_position": self.symbols.index(ticker_with_largest_cagr),
        }

    @property
    def _min_ratio_asset(self) -> dict:
        """
        The asset with the minimum ratio between the CAGR (Compound Annual Growth Rate)
        and the risk for assets that are "to the left"
        of the portfolio with the maximum CAGR on the efficiency frontier.
        """
        cagr = helpers.Frame.get_cagr(self.assets_ror)
        risk_monthly = self.assets_ror.std()
        mean_return = self.assets_ror.mean()
        risk = helpers.Float.annualize_risk(risk_monthly, mean_return)

        global_max_cagr = self.global_max_return_portfolio["CAGR"]
        global_max_risk = self.global_max_return_portfolio["Risk"]

        cagr_diff = global_max_cagr - cagr
        risk_diff = global_max_risk - risk

        if risk_diff is not None and (risk_diff == 0).any():
            risk_diff += 0.0001  # to avoid division by zero

        ratio = cagr_diff / risk_diff
        left_assets = risk_diff > 0

        if left_assets.any():
            valid_ratios = ratio[left_assets]
        else:
            right_assets = risk_diff < 0
            valid_ratios = ratio[right_assets]

        min_ticker = valid_ratios.idxmin()
        return {
            "min_asset_cagr": cagr[min_ticker],
            "ticker_with_smallest_ratio": min_ticker,
            "list_position": self.assets_ror.columns.get_loc(min_ticker),
        }

    @property
    def _max_ratio_asset_right_to_max_cagr(self) -> Optional[dict]:
        """
        The asset with the maximum ratio between CAGR and risk among assets that are to the right
        of the portfolio with the maximum CAGR on the Efficient Frontier.
        """
        cagr = helpers.Frame.get_cagr(self.assets_ror)
        risk_monthly = self.assets_ror.std()
        mean_return = self.assets_ror.mean()
        risk = helpers.Float.annualize_risk(risk_monthly, mean_return)
        tolerance = 0.01

        global_max_cagr = self.global_max_return_portfolio["CAGR"]
        global_max_risk = self.global_max_return_portfolio["Risk"]
        # TODO: global_max_cagr_is_not_asset must be TRUE if CAGR difference is small but RISK difference is big
        global_max_cagr_is_not_asset = (cagr < global_max_cagr * (1 - tolerance)).all()
        if global_max_cagr_is_not_asset:
            cagr_diff = cagr - global_max_cagr
            risk_diff = risk - global_max_risk

            if risk_diff is not None and (risk_diff == 0).any():
                risk_diff += 0.0001  # to avoid division by zero

            ratio = cagr_diff / risk_diff
            right_assets = risk_diff > 0

            if right_assets.any():
                valid_ratios = ratio[right_assets]
                max_ticker = valid_ratios.idxmax()
                return {
                    "max_asset_cagr": cagr[max_ticker],
                    "ticker_with_largest_cagr": max_ticker,
                    "list_position": self.assets_ror.columns.get_loc(max_ticker),
                }
        return None

    @property
    def _max_annual_risk_asset(self) -> dict:
        """
        Find an asset with max annual risk.
        """
        max_risk = self.risk_annual.iloc[-1, :].max()
        ticker_with_largest_risk = self.risk_annual.iloc[-1, :].nlargest(1, keep="first").index.values[0]
        return {
            "max_annual_risk": max_risk,
            "ticker_with_largest_risk": ticker_with_largest_risk,
            "list_position": self.symbols.index(ticker_with_largest_risk),
        }

    @property
    def _target_cagr_range_left(self) -> np.ndarray:
        """
        Full range of CAGR values (from min to max).
        """
        if self.full_frontier:
            cagr_series = helpers.Frame.get_cagr(self.assets_ror)
            right_asset = self._max_ratio_asset_right_to_max_cagr
            if right_asset is not None:
                exclude_ticker = right_asset.get("ticker_with_largest_cagr")
                if exclude_ticker in cagr_series.index and len(cagr_series) > 1:
                    min_cagr = cagr_series.drop(labels=exclude_ticker).min()
                else:
                    min_cagr = cagr_series.min()
            else:
                min_cagr = cagr_series.min()
        else:
            min_cagr = self.gmv_annual_values[1]
        max_cagr = self.global_max_return_portfolio["CAGR"]
        return np.linspace(min_cagr, max_cagr, self.n_points)

    @property
    def _target_cagr_range_right(self) -> Optional[np.ndarray]:
        """
        Range of CAGR values from the Global CAGR max to the max asset cagr
        to the right of the max CAGR point (if exists).
        """
        if self._max_ratio_asset_right_to_max_cagr:
            ticker_cagr = self._max_ratio_asset_right_to_max_cagr["max_asset_cagr"]
            max_cagr = self.global_max_return_portfolio["CAGR"]
            if not np.isclose(max_cagr, ticker_cagr, rtol=1e-3, atol=1e-05):
                k = abs((self._target_cagr_range_left[0] - self._target_cagr_range_left[-1]) / (max_cagr - ticker_cagr))
                # we don't want too many points in the right range. Therefore if k < 1 n_points value is used
                number_of_points = round(self.n_points / k) + 1 if k > 1 else self.n_points
                target_range = np.linspace(max_cagr, ticker_cagr, number_of_points)
                return target_range[1:]  # skip the first point (max cagr) as it presents in the left part of the EF
        return None

    @property
    def target_risk_range(self) -> np.ndarray:
        """
        Calculate range of annualized risk values (from min risk to max risk).

        The number of values in the range is defined by 'n_points'.
        The risk is defined as standard deviation of monthly rate or returns time series.

        Returns
        -------
        numpy.ndarray
            Annualized risk values (from min risk to max risk)

        Examples
        --------
        >>> frontier = ok.EfficientFrontier(['SPY.US', 'AGG.US'])
        >>> frontier.target_risk_range
        array([0.03695845, 0.04334491, 0.04973137, 0.05611783, 0.06250429,
               0.06889075, 0.07527721, 0.08166367, 0.08805012, 0.09443658,
               0.10082304, 0.1072095 , 0.11359596, 0.11998242, 0.12636888,
               0.13275534, 0.1391418 , 0.14552826, 0.15191472, 0.15830117])
        """
        min_std = self.gmv_annual_values[0]
        ticker_with_largest_risk = self.assets_ror.std().nlargest(1, keep="first").index.values[0]
        max_std_monthly = self.assets_ror.std().max()
        mean_return = self.assets_ror.loc[:, ticker_with_largest_risk].mean()
        max_std = helpers.Float.annualize_risk(max_std_monthly, mean_return)
        return np.linspace(min_std, max_std, self.n_points)

    @property
    def ef_points(self):
        """
        Generate multi-period Efficient Frontier.

        Each point on the Efficient Frontier is a rebalanced portfolio with optimized annual risk for a given CAGR.
        In case of non-convexity along the risk axis, the second part of the chart is generated,
        where the maximum risk value is found for each point.

        Returns
        -------
        DataFrame
            Table of weights and risk/return values for the Efficient Frontier.
            The columns:

            - assets weights
            - CAGR
            - Risk (standard deviation)

            All the values are annualized.

        Examples
        --------
        >>> ls = ['SPY.US', 'GLD.US']
        >>> curr = 'USD'
        >>> y = ok.EfficientFrontier(assets=ls,
        ...                             first_date='2004-12',
        ...                             last_date='2020-10',
        ...                             ccy=curr,
        ...                             rebalancing_strategy=ok.Rebalance(period='year'),
        ...                             ticker_names=True,  # use tickers in DataFrame column names (set to False to show full asset names instead of tickers)
        ...                             n_points=20,  # number of points in the Efficient Frontier
        ...                             full_frontier=False,  # draw the frontier to the global CAGR max only
        ...                             verbose=False)  # verbose mode is False to skip progress while EF points are calculated
        >>> df_reb_year = y.ef_points
        >>> df_reb_year.head(5)
               Risk      CAGR    GLD.US    SPY.US
        0  0.159400  0.087763  0.000000  1.000000
        1  0.157205  0.088171  0.014261  0.985739
        2  0.155007  0.088580  0.028941  0.971059
        3  0.152810  0.088988  0.044079  0.955921
        4  0.150615  0.089397  0.059713  0.940287

        To compare the Efficient Frontiers of annually rebalanced portfolios with not rebalanced portfolios it's possible to draw 2 charts:
        rebalancing_strategy=ok.Rebalance(period='year') and ok.Rebalance(period='none').

        >>> import matplotlib.pyplot as plt
        >>> y.rebalancing_strategy = ok.Rebalance(period='none')
        >>> df_not_reb = y.ef_points
        >>> fig = plt.figure()
        >>> # Plot the assets points
        >>> y.plot_assets(kind='cagr')
        >>> ax = plt.gca()
        >>> # Plot the Efficient Frontier for annually rebalanced portfolios
        >>> ax.plot(df_reb_year.Risk, df_reb_year.CAGR, label='Annually rebalanced')
        >>> # Plot the Efficient Frontier for not rebalanced portfolios
        >>> ax.plot(df_not_reb.Risk, df_not_reb.CAGR, label='Not rebalanced')
        >>> # Set axis labels and the title
        >>> ax.set_title('Multi-period Efficient Frontier: 2 assets')
        >>> ax.set_xlabel('Risk (Standard Deviation)')
        >>> ax.set_ylabel('Return (CAGR)')
        >>> ax.legend()
        >>> plt.show()
        """
        if self._ef_points.empty:
            self._get_ef_points()
        return self._ef_points

    def _get_ef_points(self):
        """
        Get all the points for the Efficient Frontier running optimizer.

        If verbose=True calculates elapsed time for each point and the total elapsed time.
        """
        main_start_time = time.time()

        # left part of the EF
        def compute_left_part_of_ef(i, target_cagr):
            start_time = time.time()
            row = self.minimize_risk(target_cagr)
            end_time = time.time()
            if self.verbose:
                logger.info(f"left EF point #{i + 1}/{self.n_points} is done in {end_time - start_time:.2f} sec.")
            return row

        ef_points_records = Parallel(n_jobs=-1)(
            delayed(compute_left_part_of_ef)(i, target_cagr)
            for i, target_cagr in enumerate(self._target_cagr_range_left)
        )
        # right part of the EF
        range_right = self._target_cagr_range_right
        if range_right is not None:  # range_right can be a DataFrame. Must put an explicit "is not None"

            def compute_right_part_of_ef(i, target_cagr):
                start_time = time.time()
                row = self._maximize_risk(target_cagr)
                ef_points_records.append(row)
                end_time = time.time()
                if self.verbose:
                    logger.info(
                        f"right EF point #{i + 1}/{len(range_right)} is done in {end_time - start_time:.2f} sec, init guess: {row['init_guess']}."
                    )
                return row

            ef_points_records += Parallel(n_jobs=-1)(
                delayed(compute_right_part_of_ef)(i, target_cagr) for i, target_cagr in enumerate(range_right)
            )
        df = pd.DataFrame.from_records(ef_points_records)
        df = helpers.Frame.change_columns_order(df, ["Risk", "Mean return", "CAGR"])
        main_end_time = time.time()
        if self.verbose:
            logger.info(f"Total time taken is {(main_end_time - main_start_time) / 60:.2f} min.")
        self._ef_points = df

    @property
    def mdp_points(self) -> pd.DataFrame:
        """
        Generate Most diversified portfolios frontier for rebalanced portfolios.

        Each point on the Most diversified portfolios frontier is a rebalanced portfolio with optimized
        Diversification ratio for a given CAGR.

        The points are obtained through the constrained optimization process (optimization with bounds).
        Bounds are defined with 'bounds' property.

        Returns
        -------
        DataFrame
            Table of weights and risk/return values for the Most Diversified Portfolios Frontier.
            The columns:

            - assets weights
            - CAGR (geometric mean)
            - Risk (standard deviation)
            - Diversification ratio

            All the values are annualized.

        Examples
        --------
        >>> ls4 = ['SP500TR.INDX', 'MCFTR.INDX', 'RGBITR.INDX', 'GC.COMM']
        >>> y = ok.EfficientFrontier(assets=ls4, ccy='RUB', last_date='2021-12', n_points=20)
        >>> y.mdp_points  # print mdp weights, risk, CAGR and Diversification ratio
                Risk      CAGR  Diversification ratio  ...    MCFTR.INDX   RGBITR.INDX  SP500TR.INDX
        0   0.066040  0.092220               1.234567  ...  2.081668e-16  1.000000e+00  0.000000e+00
        1   0.064299  0.093451               1.245678  ...  0.000000e+00  9.844942e-01  5.828671e-16
        ...

        To plot the Most diversification portfolios line use the DataFrame with the points data.
        Additionally 'Plot.plot_assets()' can be used to show the assets in the chart.

        >>> import matplotlib.pyplot as plt
        >>> fig = plt.figure()
        >>> # Plot the assets points
        >>> y.plot_assets(kind='cagr')  # kind should be set to "cagr" as we take "CAGR" column from the mdp_points.
        >>> ax = plt.gca()
        >>> # Plot the Most diversified portfolios line
        >>> df = y.mdp_points
        >>> ax.plot(df['Risk'], df['CAGR'])  # we chose to plot CAGR which is geometric mean of return series
        >>> # Set the axis labels and the title
        >>> ax.set_title('Most diversified portfolios line')
        >>> ax.set_xlabel('Risk (Standard Deviation)')
        >>> ax.set_ylabel('Return (CAGR)')
        >>> plt.show()
        """
        if self._mdp_points.empty:
            target_cagrs = self._target_cagr_range_left
            rows_list = []  # Collect all rows to concatenate once at the end
            for x in target_cagrs:
                row = self.get_most_diversified_portfolio(target_return=x)
                rows_list.append(row)
            df = pd.DataFrame.from_records(rows_list)
            df = helpers.Frame.change_columns_order(df, ["Risk", "CAGR"])
            self._mdp_points = df
        return self._mdp_points

    def get_monte_carlo(self, n: int = 100) -> pd.DataFrame:
        """
        Generate random rebalanced portfolios with Monte Carlo simulation.

        Risk (annualized standard deviation) and return (CAGR) are calculated for random weights within `bounds`.

        Parameters
        ----------
        n : int, default 100
            Number of random portfolios to generate with Monte Carlo simulation.

        Returns
        -------
        DataFrame
            Table with Return (CAGR) and Risk values for random portfolios (portfolios with random asset weights).

        Examples
        --------
        >>> ls_m = ['SPY.US', 'GLD.US', 'PGJ.US', 'RGBITR.INDX', 'MCFTR.INDX']
        >>> curr_rub = 'RUB'
        >>> x = ok.EfficientFrontier(assets=ls_m,
        ...                             first_date='2005-01',
        ...                             last_date='2020-11',
        ...                             ccy=curr_rub,
        ...                             rebalancing_strategy=ok.Rebalance(period='year'),  # set rebalancing period to one year
        ...                             n_points=20,
        ...                             verbose=False)
        >>> monte_carlo = x.get_monte_carlo(n=1000)  # it can take some time ...
        >>> monte_carlo.head(5)
               CAGR      Risk
        0  0.182937  0.178518
        1  0.184915  0.172965
        2  0.154892  0.141681
        3  0.185500  0.168739
        4  0.176748  0.192657

        Monte Carlo simulation results can be plotted together with the optimized portfolios on the Efficient Frontier.

        >>> import matplotlib.pyplot as plt
        >>> df_reb_year = x.ef_points  # optimize portfolios for EF. Calculations will take some time ...
        >>> fig = plt.figure()
        >>> # Plot the assets points (optional).
        >>> x.plot_assets(kind='cagr')
        >>> ax = plt.gca()
        >>> # Plot random portfolios (Monte Carlo simulation)
        >>> ax.scatter(monte_carlo.Risk, monte_carlo.CAGR)
        >>> # Plot the Efficient Frontier
        >>> ax.plot(df_reb_year.Risk, df_reb_year.CAGR, label='Annually rebalanced')
        >>> # Set the axis labels and Title
        >>> ax.set_title('Multi-period Efficient Frontier & Monte Carlo simulation')
        >>> ax.set_xlabel('Risk (Standard Deviation)')
        >>> ax.set_ylabel('CAGR')
        >>> ax.legend()
        >>> plt.show()
        """
        weights_series = helpers.Float.get_random_weights(n, self.assets_ror.shape[1], self.bounds)
        # Portfolio risk and cagr for each set of weights using cache
        rows_list = []  # Collect all rows to create DataFrame once at the end
        for weights in weights_series:
            # Use cached portfolio return time series calculation
            portfolio_ror = self._get_portfolio_ror_ts(weights)
            risk_monthly = portfolio_ror.std()
            mean_return = portfolio_ror.mean()
            risk = helpers.Float.annualize_risk(risk_monthly, mean_return)
            cagr = helpers.Frame.get_cagr(portfolio_ror)
            row = {"Risk": risk, "CAGR": cagr}
            rows_list.append(row)
        return pd.DataFrame.from_records(rows_list)

    def plot_pair_ef(self, tickers="tickers", figsize: Optional[tuple] = None) -> Axes:
        """
        Plot Efficient Frontier for every pair of assets.

        Efficient Frontier is a set of portfolios which satisfy the condition that no other portfolio exists
        with a higher expected return but with the same risk (standard deviation of return).

        Arithmetic mean (expected return) is used for optimized portfolios.

        Parameters
        ----------
        tickers : {'tickers', 'names'} or list[str], default 'tickers'
            Annotation type for assets.
            'tickers' - assets symbols are shown in form of 'SPY.US'
            'names' - assets names are used like - 'SPDR S&P 500 ETF Trust'
            To show custom annotations for each asset pass the list of names.

        figsize : tuple[float, float], default None
            Figure size (width, height) in inches. If `None`, matplotlib default is used.

        Returns
        -------
        Axes
            Matplotlib axes with the plot.

        Notes
        -----
        At least 3 assets are required.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> ls4 = ['SPY.US', 'BND.US', 'GLD.US', 'VNQ.US']
        >>> curr = 'USD'
        >>> last_date = '2021-07'
        >>> ef = ok.EfficientFrontier(ls4, ccy=curr, last_date=last_date)
        >>> ef.plot_pair_ef()
        >>> plt.show()

        It can be useful to plot the full Efficient Frontier (EF) with optimized 4 asset portfolios
        together with the EFs for each pair of assets.

        >>> ef4 = ok.EfficientFrontier(assets=ls4, ccy=curr, n_points=100)
        >>> df4 = ef4.ef_points
        >>> fig = plt.figure()
        >>> # Plot Efficient Frontier for every pair of assets. Optimized portfolios will have 2 assets.
        >>> ef4.plot_pair_ef()  # mean return is used for optimized portfolios.
        >>> ax = plt.gca()
        >>> # Plot the full Efficient Frontier for 4 asset portfolios.
        >>> ax.plot(df4['Risk'], df4['Mean return'], color = 'black', linestyle='--')
        >>> plt.show()
        """
        if len(self.symbols) < 3:
            raise ValueError("The number of symbols cannot be less than 3")
        # self._verify_axes()
        bool_inflation = hasattr(self, "inflation")
        fig, ax = plt.subplots(figsize=figsize)
        for i in itertools.combinations(self.asset_obj_dict.values(), 2):
            sym_pair = list(i)
            index0 = self.symbols.index(sym_pair[0].symbol)
            index1 = self.symbols.index(sym_pair[1].symbol)
            bounds_pair = (self.bounds[index0], self.bounds[index1])
            ef = EfficientFrontier(
                assets=sym_pair,
                ccy=self.currency,
                first_date=self.first_date,
                last_date=self.last_date,
                inflation=bool_inflation,
                full_frontier=True,
                n_points=self.n_points,
                bounds=bounds_pair,
            ).ef_points
            ax.plot(ef["Risk"], ef["CAGR"])
        self.plot_assets(kind="cagr", tickers=tickers)
        return ax

    def plot_cml(self, rf_return: float = 0, figsize: Optional[tuple] = None):
        """
        Plot Capital Market Line (CML).

        The Capital Market Line (CML) is the tangent line drawn from the point of the risk-free asset (volatility is
        zero) to the point of tangency portfolio or Maximum Sharpe Ratio (MSR) point.

        The slope of the CML is the Sharpe ratio of the tangency portfolio.

        Parameters
        ----------
        rf_return : float, default 0
            Risk-free rate of return.

        figsize : tuple[float, float], default None
            Figure size (width, height) in inches. If `None`, matplotlib default is used.

        Returns
        -------
        Axes
            Matplotlib axes with the plot.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> three_assets = ['SPY.US', 'AGG.US', 'GLD.US']
        >>> ef = ok.EfficientFrontier(assets=three_assets, ccy='USD', full_frontier=True)
        >>> ef.plot_cml(rf_return=0.05)  # Risk-Free return is 5%
        >>> plt.show()
        """
        ef = self.ef_points
        tg = self.get_tangency_portfolio(rf_return=rf_return, rate_of_return="cagr")
        fig, ax = plt.subplots(figsize=figsize)

        ax.plot(ef.Risk, ef["CAGR"], color="black")
        ax.scatter(tg["Risk"], tg["Rate_of_return"], linewidth=0, color="green", zorder=10)
        ax.annotate(
            "MSR",
            (tg["Risk"], tg["Rate_of_return"]),
            textcoords="offset points",  # how to position the text
            xytext=(-10, 10),  # distance from text to points (x,y)
            ha="center",  # horizontal alignment can be left, right or center
        )
        # plot the line
        x, y = [0, tg["Risk"]], [rf_return, tg["Rate_of_return"]]
        ax.plot(x, y, linewidth=1)
        # set the axis size
        risk_monthly = self.assets_ror.std()
        mean_return_monthly = self.assets_ror.mean()
        risks = helpers.Float.annualize_risk(risk_monthly, mean_return_monthly)
        max_return = self.global_max_return_portfolio["CAGR"]
        min_cagr = self.get_cagr().min()
        y_bottom = min(min_cagr, rf_return)
        plot_margin = 0.10
        ax.set_ylim(y_bottom * (1 - plot_margin), max_return * (1 + plot_margin))
        ax.set_xlim(0, max(risks) * (1 + plot_margin))
        # plot the assets
        self.plot_assets(kind="cagr")
        return ax

    def plot_transition_map(self, x_axe: str = "risk", figsize: Optional[tuple] = None) -> Axes:
        """
        Plot Transition Map for optimized portfolios on the Efficient Frontier.

        Transition Map shows the relation between asset weights and optimized portfolio properties:

        - CAGR (Compound annual growth rate)
        - Risk (annualized standard deviation of return)

        Parameters
        ----------
        x_axe : {'risk', 'cagr'}, default 'risk'
            Show the relation between weights and CAGR (if 'cagr') or between weights and Risk (if 'risk').
            CAGR or Risk are displayed on the x-axis.

        figsize : tuple[float, float], default None
            Figure size (width, height) in inches. If `None`, matplotlib default is used.

        Returns
        -------
        Axes
            Matplotlib axes with the plot.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> x = ok.EfficientFrontier(['SPY.US', 'AGG.US', 'GLD.US'], ccy='USD', inflation=False)
        >>> x.plot_transition_map()
        >>> plt.show()

        Transition Map with default settings shows the relation between risk (standard deviation) and asset weights
        for optimized portfolios. The same relation for CAGR can be shown by setting `x_axe='cagr'`.

        >>> x.plot_transition_map(x_axe='cagr')
        >>> plt.show()
        """
        ef = self.ef_points
        linestyle = itertools.cycle(("-", "--", ":", "-."))
        if x_axe.lower() == "cagr":
            xlabel = "CAGR (Compound Annual Growth Rate)"
            x_axe = "CAGR"
        elif x_axe.lower() == "risk":
            xlabel = "Risk (volatility)"
            x_axe = "Risk"
        else:
            raise ValueError("x_axe parameter must be 'cagr' or 'risk'.")
        fig, ax = plt.subplots(figsize=figsize)
        for i in ef:
            if i not in (
                "Risk",
                "Mean return",
                "CAGR",
                "Weights",
                "iterations",
                "init_guess",
            ):  # select only columns with tickers
                ax.plot(ef[x_axe], ef.loc[:, i], linestyle=next(linestyle), label=i)
        ax.set_xlim(ef[x_axe].min(), ef[x_axe].max())
        ax.set_xlabel(xlabel)
        ax.set_ylabel("Weights of assets")
        ax.legend(loc="upper left", frameon=False)
        fig.tight_layout()
        return ax



from __future__ import annotations

from typing import Union, Optional, Literal

import pandas as pd
import numpy as np

import okama.common.helpers.helpers as helpers
import okama.portfolios.cashflow_strategies as cf
from okama import settings


def get_wealth_indexes_fv_with_cashflow(
    ror: Union[pd.Series, pd.DataFrame],
    portfolio_symbol: Optional[str],
    inflation_symbol: Optional[str],
    cashflow_parameters: cf.CashFlow,
    task: Literal["backtest", "monte_carlo"],
) -> Union[pd.Series, pd.DataFrame]:
    """
    Calculate wealth index Future Values (FV) for a series of returns with cash flows (withdrawals/contributions).

    Values of the wealth index correspond to the beginning of the month.
    """
    dcf_object = cashflow_parameters.parent.dcf
    dcf_object.cashflow_parameters = cashflow_parameters
    period_initial_amount = cashflow_parameters.initial_investment
    period_initial_amount_cached = period_initial_amount
    last_regular_cash_flow = 0
    amount = getattr(cashflow_parameters, "amount", None)
    if isinstance(ror, pd.DataFrame):
        portfolio_position = ror.columns.get_loc(portfolio_symbol)
    else:
        # for Series
        portfolio_position = 0
        ror = ror.to_frame()
    if cashflow_parameters.NAME == "CWD":
        drawdowns = helpers.Frame.get_drawdowns(ror.iloc[:, portfolio_position])
    cash_flow_ts = dcf_object.cashflow_parameters.time_series
    # check if iteration needed
    cashflow_iterate_condition = not (cash_flow_ts.empty or (cash_flow_ts == 0).all())
    if cashflow_iterate_condition:
        ror_cashflow_df = ror.assign(cashflow_ts=cash_flow_ts)
        ror_cashflow_df.fillna(0, inplace=True)
        n_rows = ror.shape[0]
        monthly_discount_rate = (1 + dcf_object.discount_rate) ** (1 / settings._MONTHS_PER_YEAR) - 1
        discount_factors = (1.0 + monthly_discount_rate) ** np.arange(n_rows)
        if task == "backtest":
            if dcf_object.cashflow_parameters.time_series_discounted_values:
                ror_cashflow_df.loc[:, "cashflow_ts"] = ror_cashflow_df.loc[:, "cashflow_ts"].mul(
                    discount_factors, axis=0
                )
        elif task == "monte_carlo":
            if not dcf_object.cashflow_parameters.time_series_discounted_values:
                ror_cashflow_df.loc[:, "cashflow_ts"] = ror_cashflow_df.loc[:, "cashflow_ts"].mul(
                    discount_factors, axis=0
                )
        else:
            raise ValueError(f"Unknown task: {task}. It must be 'monte_carlo' or 'backtest'")
    else:
        ror_cashflow_df = ror
        ror_cashflow_df.loc[:, "cashflow_ts"] = 0
    cash_flow_ts = ror_cashflow_df["cashflow_ts"]  # cash flow monthly time series
    periods_per_year = settings.frequency_periods_per_year[cashflow_parameters.frequency]
    if hasattr(cashflow_parameters, "indexation") and cashflow_parameters.frequency != "none":
        indexation_per_period = (1 + cashflow_parameters.indexation) ** (1 / periods_per_year) - 1
    if cashflow_parameters.frequency == "month" or cashflow_parameters.frequency == "none":
        # Fast Calculation
        s = pd.Series(dtype=float, name=portfolio_symbol)
        for n, row in enumerate(ror.itertuples()):
            date = row[0]
            r = row[portfolio_position + 1]
            if cashflow_parameters.frequency == "none":
                cashflow = 0
            elif cashflow_parameters.NAME == "fixed_amount":
                cashflow = amount * (1 + indexation_per_period) ** n
            elif cashflow_parameters.NAME == "fixed_percentage":
                cashflow = cashflow_parameters.percentage / periods_per_year * period_initial_amount
            elif cashflow_parameters.NAME == "time_series":
                cashflow = 0
            elif cashflow_parameters.NAME == "CWD":
                withdrawal_without_drawdowns = amount * (1 + indexation_per_period) ** n
                if drawdowns[date] < 0:
                    cashflow = cashflow_parameters._calculate_withdrawal_size(
                        drawdown=drawdowns[date],
                        withdrawal_without_drawdowns=withdrawal_without_drawdowns,
                    )
                else:
                    cashflow = withdrawal_without_drawdowns
            else:
                raise ValueError("Wrong cashflow strategy name value.")
            period_initial_amount = period_initial_amount * (r + 1) + cashflow + cash_flow_ts[date]
            date = row[0]
            s[date] = period_initial_amount
    elif cashflow_parameters.frequency != "month" and cashflow_parameters.frequency != "none":
        # Slow Calculation
        pandas_frequency = cashflow_parameters._pandas_frequency
        months_in_full_period = settings._MONTHS_PER_YEAR / cashflow_parameters.periods_per_year
        wealth_chunks = []  # Collect all chunks to concatenate once at the end
        for n, x in enumerate(ror_cashflow_df.resample(rule=pandas_frequency, convention="start")):
            ror_ts = x[1].iloc[:, portfolio_position]  # select ror part of the grouped data
            months_local = ror_ts.shape[0]
            period_fraction = months_local / months_in_full_period  # 1 for a full period
            cashflow_ts_local = x[1].loc[:, "cashflow_ts"].copy()
            last_date = ror_ts.index[-1]
            # CashFlow inside period (Extra withdrawals/contributions)
            if (cashflow_ts_local != 0).any():
                period_wealth_index = pd.Series(dtype=float, name=portfolio_symbol)
                for k, (date, r) in enumerate(ror_ts.items()):
                    if k == 0:
                        month_balance = period_initial_amount + cashflow_ts_local[date]
                    else:
                        month_balance = month_balance * (r + 1) + cashflow_ts_local[date]
                    period_wealth_index[date] = month_balance
            else:
                period_wealth_index = period_initial_amount * (1 + ror_ts).cumprod()
            # CashFlow END period
            if cashflow_parameters.NAME == "fixed_amount":
                cashflow_value = amount * (1 + indexation_per_period) ** n
            elif cashflow_parameters.NAME == "fixed_percentage":
                cashflow_value = cashflow_parameters.percentage / periods_per_year * period_initial_amount
            elif cashflow_parameters.NAME == "VDS":
                cashflow_value = cashflow_parameters._calculate_withdrawal_size(
                    last_withdrawal=last_regular_cash_flow if n > 0 else 0,
                    balance=period_initial_amount,
                    number_of_periods=n,
                )
            elif cashflow_parameters.NAME == "CWD":
                withdrawal_without_drawdowns = amount * (1 + indexation_per_period) ** n
                if drawdowns[last_date] < 0:
                    cashflow_value = cashflow_parameters._calculate_withdrawal_size(
                        drawdown=drawdowns[last_date],
                        withdrawal_without_drawdowns=withdrawal_without_drawdowns,
                    )
                else:
                    cashflow_value = withdrawal_without_drawdowns
            else:
                raise ValueError("Wrong cashflow_method value.")
            cashflow_value *= period_fraction  # adjust cash flow to the period length (months)
            period_final_balance = period_wealth_index.iloc[-1] + cashflow_value
            period_wealth_index.iloc[-1] = period_final_balance
            period_initial_amount = period_final_balance
            wealth_chunks.append(period_wealth_index)
        wealth_df = (
            pd.concat(wealth_chunks, sort=False)
            if wealth_chunks
            else pd.DataFrame(dtype=float, columns=[portfolio_symbol])
        )
        s = wealth_df.squeeze()
    first_date = s.index[0]
    first_wealth_index_date = first_date - 1  # set first date to one month earlie
    s.loc[first_wealth_index_date] = period_initial_amount_cached
    if inflation_symbol:
        cum_inflation = helpers.Frame.get_wealth_indexes(
            ror=ror.loc[:, inflation_symbol], initial_amount=period_initial_amount_cached
        )
        wealth_index = pd.concat([s, cum_inflation], axis="columns")
    else:
        wealth_index = s
    wealth_index.sort_index(ascending=True, inplace=True)
    return wealth_index


def get_cash_flow_fv(
    ror: Union[pd.Series, pd.DataFrame],
    portfolio_symbol: Optional[str],
    cashflow_parameters: cf.CashFlow,
    task: Literal["backtest", "monte_carlo"],
) -> Union[pd.Series, pd.DataFrame]:
    """
    Calculate cash flow future values (FV) for a series of returns according to withdrawal/contributions strategies.
    """
    dcf_object = cashflow_parameters.parent.dcf
    dcf_object.cashflow_parameters = cashflow_parameters
    period_initial_amount = cashflow_parameters.initial_investment
    last_regular_cash_flow = 0
    cs_fv = pd.Series(dtype=float, name="cash_flow_fv")
    amount = getattr(cashflow_parameters, "amount", None)
    if cashflow_parameters.NAME == "CWD":
        drawdowns = helpers.Frame.get_drawdowns(ror)
    if isinstance(ror, pd.DataFrame):
        portfolio_position = ror.columns.get_loc(portfolio_symbol)
    else:
        # for Series
        portfolio_position = 0
        ror = ror.to_frame()
    cash_flow_ts = dcf_object.cashflow_parameters.time_series
    # check if iteration needed
    cashflow_iterate_condition = not (cash_flow_ts.empty or (cash_flow_ts == 0).all())
    if cashflow_iterate_condition:
        ror_cashflow_df = ror.assign(cashflow_ts=cash_flow_ts)
        ror_cashflow_df.fillna(0, inplace=True)
        n_rows = ror.shape[0]
        monthly_discount_rate = (1 + dcf_object.discount_rate) ** (1 / settings._MONTHS_PER_YEAR) - 1
        discount_factors = (1.0 + monthly_discount_rate) ** np.arange(n_rows)
        if task == "backtest":
            if dcf_object.cashflow_parameters.time_series_discounted_values:
                ror_cashflow_df.loc[:, "cashflow_ts"] = ror_cashflow_df.loc[:, "cashflow_ts"].mul(
                    discount_factors, axis=0
                )
        elif task == "monte_carlo":
            if not dcf_object.cashflow_parameters.time_series_discounted_values:
                ror_cashflow_df.loc[:, "cashflow_ts"] = ror_cashflow_df.loc[:, "cashflow_ts"].mul(
                    discount_factors, axis=0
                )
        else:
            raise ValueError(f"Unknown task: {task}. It must be 'monte_carlo' or 'backtest'")
    else:
        ror_cashflow_df = ror
        ror_cashflow_df.loc[:, "cashflow_ts"] = 0
    cash_flow_ts = ror_cashflow_df["cashflow_ts"]  # cash flow monthly time series
    periods_per_year = settings.frequency_periods_per_year[cashflow_parameters.frequency]
    if hasattr(cashflow_parameters, "indexation") and cashflow_parameters.frequency != "none":
        indexation_per_period = (1 + cashflow_parameters.indexation) ** (1 / periods_per_year) - 1
    if cashflow_parameters.frequency == "month" or cashflow_parameters.frequency == "none":
        # Fast Calculation
        for n, row in enumerate(ror.itertuples()):
            date = row[0]
            r = row[portfolio_position + 1]
            # Calculate regular cash flow
            if cashflow_parameters.frequency == "none":
                cashflow = 0
            elif cashflow_parameters.NAME == "fixed_amount":
                cashflow = amount * (1 + indexation_per_period) ** n
            elif cashflow_parameters.NAME == "fixed_percentage":
                cashflow = cashflow_parameters.percentage / periods_per_year * period_initial_amount
            elif cashflow_parameters.NAME == "time_series":
                cashflow = 0
            elif cashflow_parameters.NAME == "CWD":
                withdrawal_without_drawdowns = amount * (1 + indexation_per_period) ** n
                if drawdowns[date] < 0:
                    cashflow = cashflow_parameters._calculate_withdrawal_size(
                        drawdown=drawdowns[date],
                        withdrawal_without_drawdowns=withdrawal_without_drawdowns,
                    )
                else:
                    cashflow = withdrawal_without_drawdowns
            else:
                raise ValueError("Wrong cashflow strategy name value.")
            # add Extra Withdrawals/Contributions
            cs_value = cashflow + cash_flow_ts[date]
            period_initial_amount = period_initial_amount * (r + 1) + cs_value
            cs_fv[date] = cs_value
    elif cashflow_parameters.frequency != "month" and cashflow_parameters.frequency != "none":
        # Slow Calculation
        pandas_frequency = settings.frequency_mapping[cashflow_parameters.frequency]
        months_in_full_period = settings._MONTHS_PER_YEAR / cashflow_parameters.periods_per_year
        cashflow_chunks = []  # Collect all chunks to concatenate once at the end
        for n, x in enumerate(ror_cashflow_df.resample(rule=pandas_frequency, convention="start")):
            ror_ts = x[1].iloc[:, portfolio_position]  # select ror part of the grouped data
            months_local = ror_ts.shape[0]
            period_fraction = months_local / months_in_full_period  # 1 for a full period
            cashflow_ts_local = x[1].loc[:, "cashflow_ts"].copy()
            last_date = ror_ts.index[-1]
            # CashFlow inside period (Extra cash flow)
            if (cashflow_ts_local != 0).any():
                period_wealth_index = pd.Series(dtype=float, name=portfolio_symbol)
                for k, (date, r) in enumerate(ror_ts.items()):
                    if k == 0:
                        month_balance = period_initial_amount
                    else:
                        month_balance = month_balance * (r + 1) + cashflow_ts_local[date]
                    period_wealth_index[date] = month_balance
            else:
                period_wealth_index = period_initial_amount * (1 + ror_ts).cumprod()
            # CashFlow END period (Regular Cash Flow)
            if cashflow_parameters.NAME == "fixed_amount":
                cashflow_value = amount * (1 + indexation_per_period) ** n
            elif cashflow_parameters.NAME == "fixed_percentage":
                cashflow_value = cashflow_parameters.percentage / periods_per_year * period_initial_amount
            elif cashflow_parameters.NAME == "VDS":
                cashflow_value = cashflow_parameters._calculate_withdrawal_size(
                    last_withdrawal=last_regular_cash_flow if n > 0 else 0,
                    balance=period_initial_amount,
                    number_of_periods=n,
                )
            elif cashflow_parameters.NAME == "CWD":
                withdrawal_without_drawdowns = amount * (1 + indexation_per_period) ** n
                if drawdowns[last_date] < 0:
                    cashflow_value = cashflow_parameters._calculate_withdrawal_size(
                        drawdown=drawdowns[last_date],
                        withdrawal_without_drawdowns=withdrawal_without_drawdowns,
                    )
                else:
                    cashflow_value = withdrawal_without_drawdowns
            else:
                raise ValueError("Wrong cashflow_method value.")
            cashflow_value *= period_fraction  # adjust cash flow to the period length (months)
            last_regular_cash_flow = cashflow_value
            period_final_balance = period_wealth_index.iloc[-1] + cashflow_value
            period_wealth_index.iloc[-1] = period_final_balance
            period_initial_amount = period_final_balance
            cashflow_ts_local.iloc[-1] += cashflow_value
            cashflow_chunks.append(cashflow_ts_local)
        cs_fv = pd.concat(cashflow_chunks, sort=False) if cashflow_chunks else cs_fv
    return cs_fv


def remove_negative_values(input_s: pd.Series) -> pd.Series:
    if not isinstance(input_s, pd.Series):
        raise TypeError("input_s must be a pd.Series")
    s = input_s.copy()
    condition = s <= 0
    try:
        survival_date = s[condition].index[0]
        s[survival_date] = 0
        s[s.index > survival_date] = np.nan
    except IndexError:
        pass
    return s


def discount_monthly_cash_flow(
    cash_flow_fv: Union[pd.Series, pd.DataFrame],
    annual_effective_discount_rate: float,
    reverse: bool = False,
) -> Union[pd.Series, pd.DataFrame]:
    number_of_months = cash_flow_fv.shape[0]
    monlthly_discount_rate = (1 + annual_effective_discount_rate) ** (1 / settings._MONTHS_PER_YEAR) - 1
    if not reverse:
        discount_factors = (1.0 + monlthly_discount_rate) ** np.arange(number_of_months)
    else:
        discount_factors = (1.0 + monlthly_discount_rate) ** np.arange(number_of_months)[::-1]
    cash_flow_pv = cash_flow_fv.div(discount_factors, axis=0)
    return cash_flow_pv



"""Various validators"""

import numbers
import operator
from typing import Optional, Any

from okama import settings


def validate_integer(
    arg_name: str,
    arg_value: Any,
    *,
    min_value: Optional[int] = None,
    max_value: Optional[int] = None,
    inclusive: bool = False,
    custom_min_message: Optional[str] = None,
    custom_max_message: Optional[str] = None,
):
    """
    Validate that `arg_value` is an integer, and optionally fall within specific bounds.

    A custom override error message can be provided when min/max bounds are exceeded.

    Parameters
    ----------
    arg_name : str
        The name of the argument (used in default error messages).
    arg_value : Any
        The value being validated.
    min_value : int, optional
        Optional, specifies the minimum value.
    max_value : int, optional
        Optional, specifies the maximum value.
    inclusive : bool, default False
        Specifies whether the bounds limits are included (inclusive). Default value in False.
    custom_min_message : str, optional
        Optional, custom message when value is less than minimum.
    custom_max_message : str, optional
        Optional, custom message when value is less than maximum.

    Returns
    -------
    None
        No exceptions raised if validation passes.

    Raises
    ------
    TypeError
        If `arg_value` is not an integer.
    ValueError
        If `arg_value` does not satisfy the bounds.
    """
    if not isinstance(arg_value, int):
        raise TypeError(f"{arg_name} must be an integer.")

    ops = {"<=": operator.le, ">=": operator.ge, ">": operator.gt, "<": operator.lt}

    if inclusive:
        operator_less = "<"
        operator_greater = ">"
    else:
        operator_less = "<="
        operator_greater = ">="

    if min_value is not None and ops[operator_less](arg_value, min_value):
        if custom_min_message is not None:
            raise ValueError(custom_min_message)
        raise ValueError(f"'{arg_name:s}' must be {operator_greater} {min_value:d}")

    if max_value is not None and ops[operator_greater](arg_value, max_value):
        if custom_max_message is not None:
            raise ValueError(custom_max_message)
        raise ValueError(f"'{arg_name:s}' must be {operator_less} {max_value:d}")


def validate_real(arg_name: str, arg_value: Any) -> None:
    """
    Validate that `arg_value` is a real number.

    Parameters
    ----------
    arg_name : str
        The name of the argument (used in default error messages).
    arg_value : Any
        The value being validated.

    Returns
    -------
    None
        No exceptions raised if validation passes.

    Raises
    ------
    TypeError
        If `arg_value` is not an integer.
    """
    if not isinstance(arg_value, numbers.Real):
        raise TypeError(f"{arg_name} should be a Real number.")


def validate_distribution(distribution: Any) -> None:
    """
    Validate that `distribution` has one of allowed values.
    """
    if distribution not in settings.distributions:
        raise ValueError(f"distribution must be in {settings.distributions}.")


def validate_distribution_parameters(distr: str, param: tuple) -> None:
    if param is None:
        return
    if not isinstance(param, (list, tuple)):
        raise ValueError(f"The parameter is neither a list nor a tuple. Its type is: {type(param)}")
    match distr:
        case "norm":
            if len(param) != 2:
                raise ValueError("The parameters for Normal distribution must be of length 2 (mu, sigma).")
        case "lognorm":
            if len(param) != 3:
                raise ValueError("The parameters for Lognormal distribution must be of length 3 (shape, loc, scale).")
        case "t":
            if len(param) != 3:
                raise ValueError("The parameters for Student's T-distribution must be of length 3 (df, loc, scale).")
            if param[0] is not None and param[0] <= 2:
                raise ValueError("Degrees of freedom (df) for Student's t-distribution must be > 2.")
        case _:
            raise ValueError("Unknown distribution.")
